from InCli.SFAPI import utils,query,restClient,file
import sys, os
#python3 -m unittest

import unittest
from InCli import InCli
#from InCli.SFAPI import restClient
import traceback,time

#/Users/uormaechea/Documents/Dev/python/InCliLib/InCLipkg
# python3 -m build
#python3 -m twine upload --repository pypi dist/InCli-0.0.29*

class Test_Main(unittest.TestCase):

    def test_q(self):
        try:
            InCli._main(["-u","NOSDEV","-q", "select fields(all) from Order limit 1"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_q_System(self):
        try:
            InCli._main(["-u","NOSDEV","-q", "select fields(all) from Order limit 1","-system"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)
    def test_q_nulls(self):
        try:
            InCli._main(["-u","NOSDEV","-q", "select fields(all) from Order limit 1","-null"])       
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2) 
    def test_q_all(self):
        try:
            InCli._main(["-u","NOSDEV","-q", "select fields(all) from Order limit 1","-all"])   
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)
    def test_q_fields_all(self):
        try:
            InCli._main(["-u","NOSDEV","-q", "select fields(all) from Order limit 10","-fields","AccountId"])  
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)
    def test_q_fields_all(self):
        try:
            InCli._main(["-u","NOSDEV","-q", "select AccountId,Pricebook2Id,OrderNumber,TotalContractCost__c,State__c from Order limit 50","-fields","all"])  
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)
  
    def test_q_complexfields_all(self):
        try:
            InCli._main(["-u","NOSDEV","-q", "select Id,vlocity_cmt__BillingAccountId__c,vlocity_cmt__LineNumber__c,vlocity_cmt__ServiceAccountId__c,vlocity_cmt__Product2Id__r.productCode from orderitem where OrderId='8013O000003ivMKQAY' order by vlocity_cmt__LineNumber__c","-fields","all"])  
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

  
    def test_o(self):
        try:
            InCli._main(["-u","uormaechea.devnoscat2@nos.pt","-o"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_o_name(self):
        try:
            InCli._main(["-u","uormaechea.devnoscat4@nos.pt","-o","-name","order"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)
    def test_o_name_limit(self):
        try:
            InCli._main(["-u","uormaechea.devnoscat4@nos.pt","-o","-name","order","-limit","10"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)
    def test_o_like(self):
        try:
            InCli._main(["-u","uormaechea.devnoscat2@nos.pt","-o","-like","XOM"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_h(self):
        try:
            InCli._main(["-h"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_hu(self):
        try:
            InCli._main(["-hu"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_cc(self):
        try:
            InCli._main(["-u","DEVNOSCAT4","-cc"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_cc_no_u(self):
        try:
            InCli._main(["-cc"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)


    def test_cc_basket(self):
        try:
            InCli._main(["-u","NOSQSM","-cc","-basket"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_cc_list(self):
        try:
            InCli._main(["-u","NOSDEV","-cc","-list"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_cc_code(self):
        try:
            InCli._main(["-u","NOSDEV","-cc","-code","DC_CAT_DEEPLINK"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_cc_account(self):
        try:
            InCli._main(["-u","NOSDEV","-cc","-account","name:unaiTest4","-code","DC_CAT_MPO_CHILD_003"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_logs(self):
        try:
            InCli._main(["-u","NOSDEV","-logs"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_logs_store(self):
        try:
            InCli._main(["-logs","-store"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_logs_store_id(self):
        try:
            InCli._main(["-logs","-store","07L0Q00000N7AHsUAN"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_logs_store_error(self):
        try:
            InCli._main(["-u","NOSDEV","-logs","-store","-error"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_logs_user(self):
        try:
            res = InCli._main(["-u","NOSDEV","-logs","-loguser","Alias:ana.r"])
            print()
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_logs_query(self):
        try:
            res = InCli._main(["-u","NOSDEV","-logs","-where","Operation='Batch Apex'","-last","10"])
            print()
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_logs_userDefault_loguser(self):
        try:
            InCli._main(["-default:set","loguser","Alias:ana.r"])
            InCli._main(["-u","NOSDEV","-logs"])
            InCli._main(["-default:del","loguser"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_logs_userWrong(self):
        try:
            res = InCli._main(["-u","NOSDEV","-logs","-loguser","Alias:xxxx"])
        except Exception as e:
            print(e.args[0]['error'])
            self.assertTrue(e.args[0]['error']=='User with field Alias:xxxx does not exist in the User Object.')
            utils.printException(e)

    def test_logs_limit(self):
        try:
            res = InCli._main(["-u","NOSDEV","-logs","-limit","2"])
            self.assertTrue(len(res)==2)
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_logs(self):
        try:
            res = InCli._main(["-u","DTI","-logs"])
            if len(res)>0:
                self.assertTrue(res[0]['Id']!=None)
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)
   
    def test_logs_level(self):
        try:
            res = InCli._main(["-u","NOSDEV","-logs","07L3O00000DjhX1UAJ","-level","1"])
            print()
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_log_ID_2(self):
        try:
            id = '07L3O00000Dic7oUAB'
           # id = '07L7a00000TOMLDEA5'
            env = 'NOSDEV'
            #id = None

            if id == None:
                restClient.init(env)
                id = query.queryField("Select Id FROM ApexLog order by StartTime desc limit 1")
            InCli._main(["-u",env,"-logs",id])

        except Exception as e:
            print(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_log_limitinfo(self):
        try:
           # id = '07L3O00000Dh69HUAR'
            id = None

            if id == None:
                restClient.init('NOSDEV')
                id = query.queryField("Select Id FROM ApexLog order by StartTime desc limit 1")
            InCli._main(["-u","NOSDEV","-logs",id])
            InCli._main(["-u","NOSDEV","-logs",id,"-limitinfo"])

        except Exception as e:
            print(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)
    def test_log_IDflow(self): #flow
        try:
            InCli._main(["-u","NOSDEV","-logs","-inputfile","/Users/uormaechea/Documents/Dev/python/InCliLib/InCLipkg/test/files/flowAndWF.log"])
        except Exception as e:
            print(e)
            print(traceback.format_exc())

#07L3O00000Dy7SAUAZ
    def test_log_ID(self):
        try:
            InCli._main(["-u","demo240","-logs","07LDm000000dEFpMAM"])
        except Exception as e:
            print(e)
            print(traceback.format_exc())

    def test_log_ID_limitsInfo(self):
        try:
            InCli._main(["-u","NOSDEV","-logs","07L3O00000DxlaLUAR","-limitinfo"])
        except Exception as e:
            print(e)
            print(traceback.format_exc())
    def test_log_ID_to_file(self):
        try:
            InCli._main(["-u","DTI","-logs","07L0Q00000N5yVKUAZ","-file"])
        except Exception as e:
            print(e)
            print(traceback.format_exc())

    def test_log_folder(self):
        try:
            res = InCli._main(["-logs","-folder",f"/Users/uormaechea/Downloads/logs_SF"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_log_folder_Id(self):
        try:
            res = InCli._main(["-logs","-folder",f"/Users/uormaechea/Downloads/logs_SF"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    
    def test_log_file(self):
        try:
            res = InCli._main(["-logs","-inputfile",f"/Users/uormaechea/Documents/Dev/python/InCliLib/InCLipkg/incli/logs/07L3O00000E6dDKUAZ.log"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_log_file_to_file(self):
        try:
            res = InCli._main(["-logs","-inputfile",f"/Users/uormaechea/Documents/Dev/python/InCliLib/InCLipkg/incli/logs/07L3O00000E6dDKUAZ.log","-file"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_log_file_id(self):
        try:
            id = '07L0Q00000N5yVKUAZ'
            res = InCli._main(["-logs","-inputfile",f"/Users/uormaechea/Documents/Dev/python/InCliLib/InCLipkg/.InCli/logs/{id}.log"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_log_file_set_id(self):
        try:
            id = '07L3O00000DwvxvUAB'
            res = InCli._main(["-u","NOSDEV","-logs","-inputfile",f"/Users/uormaechea/Documents/Dev/python/InCliLib/InCLipkg/.InCli/logs/{id}.log"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_log_file_test(self):
        try:
            id = '07L0Q00000N7AHsUAN'
            res = InCli._main(["-u","NOSDEV","-logs","-inputfile",f"/Users/uormaechea/Documents/Dev/python/InCliLib/InCLipkg/.InCli/logs/{id}.log"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_log_file_to_file(self):
        try:
            restClient.init('DTI')
            html_name = f"{restClient.logFolder()}07L3O00000Dgt4sUAB.html"
            txt_name = f"{restClient.logFolder()}07L3O00000Dgt4sUAB.html"

            if file.exists(html_name):
                file.delete(html_name)
            if file.exists(txt_name):
                file.delete(txt_name)
            self.assertTrue(file.exists(html_name)==False)
            res = InCli._main(["-u","NOSDEV","-logs","-inputfile",f"/Users/uormaechea/Documents/Dev/python/InCliLib/InCLipkg/test/files/07L3O00000Dgt4sUAB.log","-file"])
            self.assertTrue(file.exists(html_name)==True)
            self.assertTrue(file.exists(txt_name)==True)

            html = file.read(f"{restClient.logFolder()}07L3O00000Dgt4sUAB.html")
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_log_file_2(self):
        try:
            InCli._main(["-u","NOSDEV","-logs","-inputfile",f"/Users/uormaechea/Documents/Dev/python/InCliLib/InCLipkg/test/files/07L3O00000DgwlbUAB.log"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_log_file_exception_wf(self):
        try:
            res = InCli._main(["-u","NOSDEV","-logs","-inputfile",f"/Users/uormaechea/Documents/Dev/python/InCliLib/InCLipkg/test/files/ExceptionThrown.log"])
            self.assertTrue(res['exception']==True)
            last = res['debugList'][-1]
            self.assertTrue(last['cmtSOQLQueries'][0] == '43')
            self.assertTrue(last['CPUTime'][0] == '13363')
            self.assertTrue(res['file_exception']==True)
            
            print()
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_log_last(self):
        try:
            InCli._main(["-u","NOSDEV","-logs","-last","100"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_log_tail_loguser_guest_auto_delete(self):
        try:
            InCli._main(["-u","demo240","-logs","-tail","-auto","-deletelogs"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_log_last_loguser(self):
        try:
            InCli._main(["-u","NOSQSM","-logs","-last","2","-loguser","username:uormaechea@nos.pt"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)
   
    def test_log_tail_auto(self):
        try:
            #return
            InCli._main(["-u","devupg2","-logs","-tail","-auto"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)  

    def test_log_tail(self):
        try:
            #return
            InCli._main(["-u","NOSQSM","-logs","-tail"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)   
    def test_log_tail_delete_guest(self):
        try:
            #return
            InCli._main(["-u","NOSQSM","-logs","-tail","-deletelogs"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)     

    def test_log_tail_delete(self):
        try:
            #return
            InCli._main(["-u","NOSQSM","-logs","-tail","-deletelogs"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)     
    def test_log_tail_where(self):
        try:
            return

            InCli._main(["-u","NOSDEV","-logs","-tail","-where","LogLength>3000"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)    
    def test_log_tailx(self):
        try:
            restClient.init('NOSDEV')
            logRecords = query.queryRecords("Select fields(all) FROM ApexLog order by StartTime desc limit 1")
            time = logRecords[0]['StartTime']
            timez = time.split('.')[0] + "Z"
            while (True):
                logRecords = query.queryRecords(f"Select Id,LogUserId,LogLength,LastModifiedDate,Request,Operation,Application,Status,DurationMilliseconds,StartTime,Location,RequestIdentifier FROM ApexLog where StartTime > {timez} order by StartTime asc ")

                if len(logRecords) > 0:
                    print()
                    for record in logRecords:
                        print(f"{record['StartTime']}  {record['Operation']}")
                        time = record['StartTime']
                        timez = time.split('.')[0] + "Z"
                        
                time.sleep(5)
            print()

          #  InCli._main(["-u","NOSDEV","-logs","-last","10"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_default(self):
        try:
            InCli._main(["-default:set","u"])
            InCli._main(["-default:set","u","NOSDEV"])
            InCli._main(["-default:get","u"])        
            res = InCli._main(["-logs","-last","1"])
            InCli._main(["-default:del","u"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_default_get(self):
        try:
            InCli._main(["-default:get"])        
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_default_weird(self):
        try:
            InCli._main(["-h:","InCli","-default:set","u","NOSDEV"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_d(self):
        try:
            InCli._main(["-u","NOSDEV","-d"])
            InCli._main(["-u","NOSDEV","-d","Order"])
            InCli._main(["-u","NOSDEV","-d","Order:Status"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_l(self):
        try:
            InCli._main(["-u","NOSPRD","-l"])
            InCli._main(["-u","DEVNOSCAT2","-l"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_l_no_U(self):
        try:
            InCli._main(["-l"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_checkVersion(self):
        InCli.checkVersion()
        print()

    def test_ipe(self):
        try:
            InCli._main(["-u","NOSDEV","-ipe"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_ipe_id(self):
        try:
            InCli._main(["-u","NOSDEV","-ipe","a6K3O000000FJ5WUAW"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_ipe_id_tofile(self):
        try:
            InCli._main(["-u","NOSDEV","-ipe","a6K3O000000FJ5WUAW","-file"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)
    def test_ipe_last(self):
        try:
            InCli._main(["-u","NOSDEV","-ipe","-last","2"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_ipa(self):
        try:
            InCli._main(["-u","NOSDEV","-ipa"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_ipa_id(self):
        try:
            InCli._main(["-u","NOSQSM","-ipa","a2J7a000002UakuEAC"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_ipa_id_toFile(self):
        try:
            InCli._main(["-u","NOSQSM","-ipa","a2J7a000002UakuEAC","-file"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_vte(self):
        try:
            InCli._main(["-u","NOSDEV","-vte"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_vte(self):
        try:
            InCli._main(["-u","NOSDEV","-vte","a6a3O000000FFkYQAW"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)


from . import utils,Sobjects

def setTimes(context,line,obj=None,field=None,value=None,chunkNum=None,type=None):
    def addList(obj,field,val):
        if field in obj:
            obj[field].append(val)
        else:
            obj[field] = [val]

    chunks = line.split('|')

    if obj == None:
        obj = {
            'type' : type,
            'ident' : context['ident'],
            'exception' :False
        }
       
        if len(chunks)>3:  obj['Id'] = chunks[3]

    addList(obj,'lines',line)
    addList(obj,'CPUTime',context['DEF:CPU time'])
    addList(obj,'SOQLQueries',context['DEF:SOQL queries'])
    addList(obj,'cmtCPUTime',context['CMT:CPU time'])
    addList(obj,'cmtSOQLQueries',context['CMT:SOQL queries'])
    addList(obj,'totalQueries',context['totalQueries'])
    addList(obj,'time',chunks[0].split(' ')[0])
    if len(chunks)>1:  addList(obj,'timeStamp',int ((chunks[0].split('(')[1]).split(')')[0]))
    else:  addList(obj,'timeStamp',0)

    if obj['type'] is None:
        print()

    if field is not None:  obj[field] = chunks[chunkNum] if value is None else value

    if context['timeZero'] == 0:  context['timeZero'] = obj['timeStamp'][0]

    obj['elapsedTime'] = obj['timeStamp'][0] #- _context['timeZero']

    return obj

def append_and_increaseIdent(context,obj,increase=True):
    context['openItemsList'].append(obj)
    if increase == True: context['ident'] = context['ident'] + 1
    context['parsedLines'].append(obj)

def decreaseIdent_pop_setFields(context,type,value,key='key',endsWith=None,decrease=True):
    obj = popFromList(context,type=type,key=key,value=value,endsWith=endsWith)
    if obj == None:
        a=1
    else:
        if decrease == True:   context['ident'] = obj['ident']
        setTimes(context,context['line'],obj)
    return obj

def getFromDebugList(context,values):
    for line in reversed(context['parsedLines']):
        for key in values.keys():
            if key not in line:
                break
            if line[key]!=values[key]:
                break
        return line
    return None    

def popFromList(context,type,value,key='key',endsWith=False):
    openItemsList = context['openItemsList']
    try:
        #for i,obj in enumerate(openItemsList):

        for i,obj in reversed(list(enumerate(openItemsList))):
            if obj['type'] == type:
                if endsWith == True:
                    if key not in obj:      continue
                    if obj[key].endswith(value) or obj[key].startswith(value):
                        openItemsList.pop(i)
                        return obj    
                else:
                    if key not in obj:     continue
                    if obj[key] == value:
                        openItemsList.pop(i)
                        return obj
    except Exception as e:    print(e) 

    return None           

def getFromList(theList,field,value,endsWith=False,delete=True,startsWith=False):
    try:
        rvs = theList
        for i,obj in reversed(list(enumerate(rvs))):
            if field in obj:
                if startsWith == True:
                    if obj[field].startswith(value):
                        if delete == True:
                            rvs.pop(i)
                        return obj    
                if endsWith == True:
                    if obj[field].endswith(value):
                        if delete == True:
                            rvs.pop(i)
                        return obj    
                else:
                    if obj[field] == value:
                        if delete==True:
                            rvs.pop(i)
                        return obj
    except Exception as e:
        print(e) 
    return None

def is_in_operation(context,text,contains=False):
    if context['chunks_lenght']<2: return False
    if contains and text in context['chunks'][1]: return True
    elif context['chunks'][1] == text: return True
    return False

def parseWfRule(context):
    line = context['line']
    chunks = context['chunks'] 

    if is_in_operation(context,'WF_RULE_EVAL',contains=True):
        if 'BEGIN' in chunks[1]:
            obj = setTimes(context,line,field='output',value='Workflow',type='RULE_EVAL')
            append_and_increaseIdent(context,obj)
            return True

        if 'END' in chunks[1]:
            decreaseIdent_pop_setFields(context,type='RULE_EVAL',key='output',value='Workflow')
            return True

    if is_in_operation(context,'WF_CRITERIA',contains=True):
        if 'BEGIN' in chunks[1]:
            obj = setTimes(context,line,type='WF_CRITERIA')

            colon_split=chunks[2].split(':')
            colon_space = colon_split[1].strip().split(' ')
            obj['ObjectName'] = colon_split[0][1:]
            obj['RecordName'] = colon_space[0]
            if len(colon_space)>1:
                obj['RecordID'] = colon_space[1]
            else:
                obj['RecordID'] = ""

            obj['rulename'] = chunks[3]
            obj['rulenameId'] = chunks[4]
            obj['output'] = obj['rulename']

            append_and_increaseIdent(context,obj)
            return True

        if 'END' in chunks[1]:
            obj =decreaseIdent_pop_setFields(context,type='WF_CRITERIA',key='type',value='WF_CRITERIA')   
            obj['result'] = chunks[2]
            obj['output'] = f"{obj['ObjectName']}: {obj['rulename']} --> {obj['result']}"
            return True
  
    if is_in_operation(context,'WF_RULE_NOT_EVALUATED'):
        obj =decreaseIdent_pop_setFields(context,type='WF_CRITERIA',key='type',value='WF_CRITERIA')   
        obj['output'] = f"{obj['rulename']} --> Rule Not Evaluated"
        return True

    if is_in_operation(context,'WF_ACTION'):
        obj = getFromList(context['openItemsList'],'output','Workflow',delete=False)
        obj['action'] = chunks[2]
        return True

def parseExceptionThrown(context):
    line = context['line']
    chunks = context['chunks']

  #  if context['chunks_lenght']>1 and chunks[1] == 'EXCEPTION_THROWN':
    if is_in_operation(context,'EXCEPTION_THROWN'):
        obj = setTimes(context,line,type='EXCEPTION',field='output',value=chunks[3])
        context['exception'] = True
        context['exception_msg'] = obj['output']

        context['parsedLines'].append(obj)
        context['file_exception'] = True
        next = 1
        nextline = context['lines'][context['line_index']+next]
        while '|' not in nextline:
            if nextline != '':
                obj = context['parsedLines'][-1].copy()
                context['parsedLines'].append(obj)
                obj['output'] = nextline
            next = next + 1
            nextline = context['lines'][context['line_index']+next]
        return True

    if is_in_operation(context,'FATAL_ERROR'):
   # if context['chunks_lenght']>1 and chunks[1] == 'FATAL_ERROR':
        obj = setTimes(context,line,type='EXCEPTION',field='output',value=chunks[2])
        context['exception'] = True
        context['exception_msg'] = obj['output']

        context['parsedLines'].append(obj)
        context['file_exception'] = True
        next = 1
        nextline = context['lines'][context['line_index']+next]
        while '|' not in nextline:
            if nextline != '':
                obj = context['parsedLines'][-1].copy()
                context['parsedLines'].append(obj)
                obj['output'] = nextline
            next = next + 1
            nextline = context['lines'][context['line_index']+next]
        return True

    return False

def parseUserDebug(context):
    line = context['line']
    chunks = context['chunks']

    if is_in_operation(context,'USER_DEBUG'):
        obj = setTimes(context,line,type='DEBUG')
        obj['timeStamp'].append(obj['timeStamp'][0])
        obj['type'] = 'DEBUG'
        obj['subType'] = chunks[3]
        obj['output'] = chunks[4]
        if obj['subType'] == 'ERROR':
            context['exception'] = True
            context['exception_msg'] = obj['output']
        obj['apexline'] = chunks[2][1:-1]
        context['parsedLines'].append(obj)
        if context['line_index']<(len(context['lines'])-1):
            next = 1
            nextline = context['lines'][context['line_index']+next]
            while '|' not in nextline:
                obj = context['parsedLines'][-1].copy()
                context['parsedLines'].append(obj)
            #    obj['string'] = nextline
                obj['output'] = nextline
                next = next + 1
                nextline = context['lines'][context['line_index']+next]

        if  obj['output'].startswith('*** getCpuTime() ***'):
            chs = chunks[4].split(' ')
            context[f'DEF:CPU time'] = chs[4]
        if  obj['output'].startswith('CPU Time:'):
            chs = chunks[4].split(' ')
            context[f'DEF:CPU time'] = chs[2]  
        if  obj['output'].startswith('*** getQueries() ***'):  
            chs = chunks[4].split(' ')
            context[f'DEF:SOQL queries'] = chs[4]
        if obj['output'].startswith('CPQCustomHookImplementation'):
            if obj['output'].endswith('PreInvoke'):
                context['CPQCustomHookImplementation'] = 'Started'
            if obj['output'].endswith('PostInvoke'):
                context['CPQCustomHookImplementation'] = 'Finished'

        return True

    return False

def parseLimits(context):
    line = context['line']
    chunks = context['chunks'] 

    if is_in_operation(context,'LIMIT_USAGE'):
        if chunks[3] == 'SOQL':
            context[f'DEF:SOQL queries'] = chunks[4]
        return True

    if '|LIMIT_USAGE|' in line and '|SOQL|' in line: 
        context[f'DEF:SOQL queries'] = chunks[4]
        return True

    if is_in_operation(context,'LIMIT_USAGE_FOR_NS'):
        obj = setTimes(context,line,type='LIMIT')
        obj['output'] = f"{chunks[1].lower()}  {chunks[2]}"
        context['parsedLines'].append(obj)

        limits = chunks[2]
        if limits == '(default)':         limitsNS = 'DEF:'
        elif limits == 'vlocity_cmt':     limitsNS = 'CMT:'
        else:                             limitsNS = f"{limits}:"

        next = 1
        nextline = context['lines'][context['line_index']+next]
        while '|' not in nextline:
            if 'SOQL queries' in nextline:
                nlchunks = nextline.split(' ')
                if f'{limitsNS}SOQL queries' not in context:
                    context[f'{limitsNS}SOQL queries'] = 0
                if int(context[f'{limitsNS}SOQL queries']) < int(nlchunks[6]):
                    context[f'{limitsNS}SOQL queries'] = nlchunks[6]
            if 'CPU time' in nextline:
                nlchunks = nextline.split(' ')
                if f'{limitsNS}CPU time' not in context:
                    context[f'{limitsNS}CPU time'] = 0
                if int(context[f'{limitsNS}CPU time']) < int(nlchunks[5]):
                    context[f'{limitsNS}CPU time'] = nlchunks[5]
            next = next + 1
            nextline = context['lines'][context['line_index']+next]
        return True

def parseSOQL(context):
    line = context['line']
    chunks = context['chunks']
    if is_in_operation(context,'SOQL_EXECUTE_BEGIN'):
        obj = setTimes(context,line,type="SOQL")
        obj['query'] = chunks[4]
        obj['object'] = chunks[4].lower().split(' from ')[1].strip().split(' ')[0]
        obj['apexline'] = chunks[2][1:-1]

        soql = obj['query'].lower()
        ch_so = obj['query'].split("'")
        if len(ch_so)>1:
            posibles = ch_so[1::2]

            ids = [posible for posible in posibles if Sobjects.checkId(posible) ]
            idss = set(ids)
            if len(idss)>0:
                obj['where_ids'] = ",".join(idss)


        obj['for_update'] = ' for update' in soql

        if 'where' in soql:
            soql = soql.split('where')[0]
        _from = soql.split(' from ')[-1].strip()
        _from = _from.split(' ')[0]

        obj['from'] = _from
        obj['output'] = f"Select: {obj['from']} --> No SOQL_EXECUTE_END found"

        append_and_increaseIdent(context,obj,increase=False)
        return True

    if context['chunks_lenght']>1 and chunks[1] == 'SOQL_EXECUTE_END':
        context['totalQueries'] = context['totalQueries'] + 1
        obj = decreaseIdent_pop_setFields(context,type="SOQL",key='type',value='SOQL',decrease=False)
        obj['rows'] = chunks[3].split(':')[1]
        for_uptate = "for update" if obj['for_update'] else ""
        ids = f"w:{obj['where_ids']}" if 'where_ids' in obj else ""
        obj['output'] = f"Select {for_uptate}: {obj['from']} --> {obj['rows']} rows {utils.CFAINT}{utils.CYELLOW}{ids}{utils.CEND}"

        return True

    return False

def parseMethod(context):
    line = context['line']
    chunks = context['chunks'] 
    if context['chunks_lenght']>1 and 'METHOD_' in  chunks[1]:
        if len(chunks)<4:
            print(line)
            return

        operation = chunks[1]
       ## method = getMethod(line)
        method = chunks[3] if len(chunks) == 4 else chunks[4]
        if '(' in method:
            method = method.split('(')[0]

        if 'ENTRY' in operation:
            obj = setTimes(context,line,type='METHOD')
            obj['method'] = method
            obj['apexline'] = chunks[2][1:-1] if chunks[2]!='[EXTERNAL]' else 'EX'
            obj['output'] = obj['method']
            context['parsedLines'].append(obj)

            if '.getInstance' in obj['method']:
                pass
            else:
                context['openItemsList'].append(obj)
                context['ident'] = context['ident'] + 1
            return True

        else:
            obj = getFromList(context['openItemsList'],'method',method)
            if obj == None:
                obj = getFromList(context['openItemsList'],'method',f"{method}",endsWith=True)
                apexline = chunks[2][1:-1]
                if obj != None and apexline != obj['apexline']:
                    obj == None
            if obj == None:
                obj = getFromList(context['openItemsList'],'method',f"{method}",startsWith=True)
                apexline = chunks[2][1:-1]
                if obj != None and apexline != obj['apexline']:
                    obj == None
            if obj is not None:
                context['ident'] = obj['ident']
                setTimes(context,line,obj)

            else:
                obj = setTimes(context,line,type='NO_ENTRY')
                obj['method'] = chunks[-1]
                obj['apexline'] = chunks[2][1:-1] if chunks[2]!='[EXTERNAL]' else 'EX'
                context['parsedLines'].append(obj)

            if 'method' in obj:
                obj['output']=obj['method']
            else:
                obj['output']=obj['Id']
            return True

    return False

def parseVariableAssigment(context):
    line = context['line']
    chunks = context['chunks'] 

    if 'EXP_VAR' in context and context['EXP_VAR'] == True:
        if chunks[1] == 'VARIABLE_ASSIGNMENT' and chunks[2] == '[EXTERNAL]':
            obj = setTimes(context,line,type='VAR_ASSIGN')
            obj['type'] = 'VAR_ASSIGN'
            obj['subType'] = 'EXCEPTION'
         #   obj['string'] = chunks[4]
            obj['apexline'] = chunks[2][1:-1] if chunks[2]!='[EXTERNAL]' else 'EX'

            context['parsedLines'].append(obj)         
            obj['output'] = chunks[4]

        else:   context['EXP_VAR'] = False
        return False #why False???

    if is_in_operation(context,'VARIABLE_ASSIGNMENT'):
        if len(chunks) >= 5:
            if 'ExecutionException' in chunks[4] or 'ExecutionException' in chunks[4]:
                obj = setTimes(context,line,type='VAR_ASSIGN')
                obj['type'] = 'VAR_ASSIGN'
                obj['subType'] = 'EXCEPTION'
                obj['string'] = chunks[4]
                obj['apexline'] = chunks[2][1:-1] if chunks[2]!='[EXTERNAL]' else 'EX'

                context['parsedLines'].append(obj)
                obj['output'] = obj['string'] 

                context['EXP_VAR'] = True
        return True
    return False

def parseDML(context):
    line = context['line']
    chunks = context['chunks']

    if is_in_operation(context,'DML_BEGIN'):
        obj = setTimes(context,line,type="DML")
        obj['OP'] = chunks[3]
        obj['Type'] = chunks[4]
        obj['Id'] = chunks[2]
        obj['Rows'] = chunks[5]
        obj['apexline'] = chunks[2][1:-1]
        obj['output'] = f"{obj['OP']} {obj['Type']} --> {obj['Rows']}" 
        append_and_increaseIdent(context,obj)
        return True

    if is_in_operation(context,'DML_END'):
        decreaseIdent_pop_setFields(context,'DML',key='Id',value=chunks[2])
        return True

    return False

def parseCallOutResponse(context):
    line = context['line']
    chunks = context['chunks']

    if is_in_operation(context,'CALLOUT_RESPONSE'):
        obj = setTimes(context,line,type='CALLOUT')
     #   obj['string'] = chunks[3]
        obj['apexline'] = chunks[2][1:-1]

        context['parsedLines'].append(obj)  
        obj['output'] = chunks[3]
        return True

    return False

def parseConstructor(context):
    line = context['line']
    chunks = context['chunks']

    if is_in_operation(context,'CONSTRUCTOR_ENTRY'):
        obj = setTimes(context,line,field='output',value=chunks[5],type='CONSTRUCTOR')
        obj['apexline'] = chunks[2][1:-1] if chunks[2]!='[EXTERNAL]' else 'EX'

        append_and_increaseIdent(context,obj)
        return True

    if is_in_operation(context,'CONSTRUCTOR_EXIT'):

        decreaseIdent_pop_setFields(context,type='CONSTRUCTOR',key='output',value=chunks[5])
        return True

    return False

def parseCodeUnit(context):
    line = context['line']
    chunks = context['chunks']

    if is_in_operation(context,'CODE_UNIT_STARTED'):
        obj = setTimes(context,line,type='CODE_UNIT')
        obj['output'] = chunks[4] if len(chunks)>4 else chunks[3]
        append_and_increaseIdent(context,obj)
        return True

    if is_in_operation(context,'CODE_UNIT_FINISHED'):
        decreaseIdent_pop_setFields(context,'CODE_UNIT',key='output',value=chunks[2])
        return True

    return False

def parseNamedCredentials(context):
    line = context['line']
    chunks = context['chunks']

    if is_in_operation(context,'NAMED_CREDENTIAL_REQUEST'):
        obj = setTimes(context,line,field='output',value=chunks[2],type='NAMED_CRD')
        append_and_increaseIdent(context,obj)
        return True

    if is_in_operation(context,'NAMED_CREDENTIAL_RESPONSE'):
        obj = decreaseIdent_pop_setFields(context,type='NAMED_CRD',key='type',value='NAMED_CRD')
        return True

    return False

def parseFlow(context):
    line = context['line']
    chunks = context['chunks']
    debugList = context['parsedLines']

    if 1==2:
        if '|FLOW_START_INTERVIEWS_BEGINxx|' in line:
            obj = setTimes(context,line,type='FLOW_START_INTERVIEWS',field='output',value='FLOW_START_INTERVIEWS')
            append_and_increaseIdent(context,obj)

        if '|FLOW_START_INTERVIEWS_ENDxx|' in line:
            decreaseIdent_pop_setFields(context,'FLOW_START_INTERVIEWS',key='output',value='FLOW_START_INTERVIEWS')

    if is_in_operation(context,'FLOW_START_INTERVIEW_BEGIN'):
        obj = setTimes(context,line,type='FLOW_START_INTERVIEW')
        obj['interviewId'] = chunks[2]
        obj['Name'] = chunks[3]
        obj['output'] = obj['Name']
        append_and_increaseIdent(context,obj)
        return True

    if is_in_operation(context,'FLOW_START_INTERVIEW_END'):
        interviewId = chunks[2]
        decreaseIdent_pop_setFields(context,'FLOW_START_INTERVIEW',key='interviewId',value=interviewId)
        return True

    if is_in_operation(context,'FLOW_ELEMENT_ERROR'):
        obj = setTimes(context,line,type='FLOW_ELEMENT_ERROR')
        obj['message'] = chunks[2]
        obj['elementType'] = chunks[3]
        obj['elementName'] = chunks[4]
        obj['output'] = utils.CRED+ f"{obj['message']} in {obj['elementType']}:{obj['elementName']}" + utils.CEND
        debugList.append(obj)
        context['exception'] = True
        context['exception_msg'] = obj['output']
        return True
    
    if is_in_operation(context,'FLOW_ELEMENT_BEGIN'):
        obj = setTimes(context,line,type='FLOW_ELEMENT')
        obj['interviewId'] = chunks[2]
        obj['elementType'] = chunks[3]
        obj['elementName'] = chunks[4]
        obj['output'] = f"{obj['elementType']}-{obj['elementName']}"
        append_and_increaseIdent(context,obj)
        return True

    if is_in_operation(context,'FLOW_ELEMENT_END'):
        interviewId = chunks[2]
        decreaseIdent_pop_setFields(context,'FLOW_ELEMENT',key='interviewId',value=interviewId)

    if is_in_operation(context,'FLOW_RULE_DETAIL'):
        values = {
            'type':'FLOW_ELEMENT',
            'elementType':'FlowDecision',
            'interviewId':chunks[2],
            'elementName':chunks[3]
        }
        obj = getFromDebugList(context,values)
        obj['ruleName'] = chunks[3]
        obj['result'] = chunks[4]
        obj['output'] = f"{obj['elementType']}-{obj['elementName']} -- {obj['ruleName']}->{obj['result']}"
        return True

    return False

def parseUserInfo(context):
    if is_in_operation(context,'USER_INFO'):
   # if '|USER_INFO|' in context['line']:
        obj = setTimes(context,context['line'],field='output',value=context['chunks'][4],type='USER_INFO')
        context['parsedLines'].append(obj)
        return True
    return False

def appendEnd(context):

    for line in reversed(context['lines']):
        if '|' in line:
            break

    if 'CPQCustomHookImplementation' in context and  context['CPQCustomHookImplementation'] == 'Started':
        obj = setTimes(context,line,type='EXCEPTION',field='output',value="CPQCustomHookImplementation did not finish")
        context['exception'] = True
        context['exception_msg'] = obj['output']

        context['parsedLines'].append(obj)
        context['file_exception'] = True
        
    lastline = line
    obj = setTimes(context,lastline,type="END")
    obj['output'] = 'Final Limits'
    context['parsedLines'].append(obj)


import unittest
from InCli import InCli
from InCli.SFAPI import file,restClient,tooling,Sobjects,utils,traceFlag,query

class Test_TraceFlag(unittest.TestCase):
    def test_descibre(self):
        restClient.init('NOSDEV')
        tooling.describe('DebugLevel')
        tooling.describe('TraceFlag')

    #    tooling.query("a")
        userId = Sobjects.IdF('User','username:uormaechea_nosdev@nos.pt')

        call = traceFlag.get_traceflags_for_user('username:uormaechea_nosdev@nos.pt')
        id = tooling.idF('TraceFlag',f"TracedEntityId:{userId}")

        tf = tooling.get('TraceFlag',id)
        utils.printFormated(tf)

        DebugLevelId = tooling.idF('DebugLevel',"DeveloperName:InCli")   

        call = tooling.idF('TraceFlag',f"DebugLevelId:{DebugLevelId}")

        q = f"select id from TraceFlag where DebugLevelId = '{DebugLevelId}' and TracedEntityId = '{userId}'"
        call = tooling.query(q)  
        traceFlagId = call['records'][0]['Id']

        traceFlag.update_trace_flag_incli(traceFlagId)
        call = tooling.get("TraceFlag","7tf3O000001LbmoQAC")
        
        call2 =  traceFlag.create_trace_flag_incli(userId=userId,DebugLevelId=DebugLevelId)
        print()

    def test_create(self):
        restClient.init('DTI')
        traceFlag.create_debug_level_incli()

    def test_set_incli_traceFlag_for_user(self):
        restClient.init('DTI')

        userF = "username:onboarding@nosdti-parceiros.cs109.force.com"
        traceFlag.set_incli_traceFlag_for_user(userF)

    def test_debuglevel_get(self):
        restClient.init('NOSDEV')

        call = tooling.tooling_idF('DebugLevel',"DeveloperName:XXXX")   
        self.assertTrue(call==None)

        id = tooling.tooling_idF('DebugLevel',"DeveloperName:InCli")   
        if id != None:
            tooling.tooling_delete('DebugLevel',id)
            id = tooling.tooling_idF('DebugLevel',"DeveloperName:InCli")   

            print()
        if id == None:
            tooling.create_debug_level_incli()
            print()

    def test_query_who(self):

        envs = ['DTI',"NOSDEV","NOSQSM","NOSPRD"]
        for env in envs:
            try:
                restClient.init(env)

                DebugLevelId = tooling.idF('DebugLevel',"DeveloperName:InCli")   
                q = f"select TracedEntityId,StartDate,ExpirationDate from TraceFlag where DebugLevelId='{DebugLevelId}' limit 100"

                res =tooling.query(q)
                q1 = f"select username from User where Id='{res['records'][0]['TracedEntityId']}'"
                res1 = query.query(q1)
                print()
                print(f'Environment: {env}')
                utils.printFormated(res1['records'])
            except Exception as e:
                utils.printException(e)

        print()



from . import  file

def read_as_lists(filepath):
    filepath = file.addExtension(filepath,".csv")

    lineslist = []
    f = open(f'{filepath}', 'r',encoding='utf-8-sig')
    lines = f.readlines()

    for line in lines:
        linelist = line.split(";")
        i = 0
        while i < len(linelist):
            linelist[i] = linelist[i].rstrip("\n")
            if linelist[i] == '':   #$$$$ Should there be a NULL?
                linelist[i] = None
            i = i +1

        lineslist.append(linelist)

    return lineslist

def read(filepath):
    filepath = file.addExtension(filepath,".csv")

    f = open(f'{filepath}', 'r',encoding='utf-8-sig')

    #f = open(f'{filepath}', 'r')
    lines = f.readlines()

    firstline = True    
    fieldsMap = []
    headers = None

    for line in lines:
        field = {}
        if firstline == True:
            headers = line.split(";")
            firstline = False
            continue

        items = line.split(";")

        i = 0
        for he in headers:
            value = items[i].rstrip("\n")
            if value == '':
                value = None
            field[he.rstrip("\n")] = value
            i = i +1

        fieldsMap.append(field)

    return fieldsMap

def write(filepath,obj):
    filepath = file.addExtension(filepath,".csv")
    f = write_open(filepath,obj,mode='w',header=True)
    write_close(f)
    return file.abspath(filepath)

def write_open(filepath,objects,mode="w",header=False):
    filepath = file.addExtension(filepath,".csv")
    f = open(filepath,mode)     

    write_objects(f,objects,header)
     
    return f

def write_objects(f,objects,header=False):
    def write_obj(obj):
        values = []
        for key in obj.keys():
            values.append(str(obj[key]))
        line = ";".join(values)
        print(line,file=f)       

    if header == True:
        if type(objects) is dict:
            keys = list(objects.keys())
        if type(objects) is list:
            keys = list(objects[0].keys())
        line = ";".join(keys)
        print(line,file=f)   

    if type(objects) is list:
        for obj in objects:
            write_obj(obj)   
    if type(objects) is dict:
        write_obj(objects)   

def write_close(f):
    f.close()


import unittest,simplejson
from InCli.SFAPI import restClient,query,Sobjects,utils,tooling


class Test_z_compareOrgs(unittest.TestCase):
    definitions = []
    def test_limits(self):
        restClient.init('NOSQSM')

        sObjectName = 'product2'
      #  sObjectName = 'Family__c'
      #  sObjectName = 'Spectrum__c'
      #  sObjectName = 'Level__c'
      #  sObjectName =  'Product__c'
        sObjectName = 'ProductCorporate__c'
        sObjectName = 'Order'
        sObjectName = 'Account'

        prod_qsm = Sobjects.describe(sObjectName)

        self.definitions = self.get_custom_fields(sObjectName)

        restClient.init('demo240')
        prod_demo240 = Sobjects.describe(sObjectName)

        prod_qsm_field_names = [p['name'] for p in prod_qsm['fields']]
        prod_demo240_field_names = [p['name'] for p in prod_demo240['fields']]

        diff_list = list (set(prod_qsm_field_names).difference(prod_demo240_field_names))
        for field in prod_qsm['fields']:
            name = field['name']
            if name not in prod_demo240_field_names:
                print('-------------------------------------------------------')
                print(name)
                print(field['type'])
                try:
                    data = self.calc_customField(sObjectName,field)
                    #data = self.customField(sObjectName,field)
                    call = self.create_custom_field(data)
                    self.set_field_permission(sObjectName,data=data)
                    print('DONE!!!!!!')
                except Exception as e:
                    if 'errorCode' in e.args[0] and e.args[0]['errorCode'] == 'DUPLICATE_DEVELOPER_NAME':
                        continue
                    utils.printException(e)


        print()

    def create_custom_field(self,data):
        action ='/services/data/v54.0/tooling/sobjects/CustomField'
        method = 'post'

        call= restClient.callAPI(action=action,method=method,data=data)
        restClient.checkError()
        print(call)
        return call

    def get_custom_field(self,sobject,field):
        data = { "FullName": f"{sobject}.{field}"}
        action ='/services/data/v54.0/tooling/sobjects/CustomField'
        method = 'get'

        call= restClient.callAPI(action=action,method=method,data=data)
        restClient.checkError()
        print(call)

    def get_field_definition(self):

        res = tooling.query(f"select fields(all) from FieldDefinition WHERE EntityDefinition.QualifiedApiName IN ('Product2') limit 100")

        res2 = tooling.query(f"SELECT NamespacePrefix, DeveloperName, TableEnumOrId,ManageableState FROM CustomField ")
        for r in res2['records']:
            print(f"{r['TableEnumOrId']}  {r['NamespacePrefix']} {r['DeveloperName']}")
            if 'Availability_SLA_Formula' == r['DeveloperName']:
                res3 = tooling.query(f"SELECT NamespacePrefix, DeveloperName, TableEnumOrId,ManageableState,Metadata FROM CustomField where DeveloperName='{r['DeveloperName']}' limit 1")

        print()

    def get_custom_fields(self,sobject,namespace=None):
        
        tableEnumOrId = sobject
        is_custom_object = False

        if '__c' in sobject:
            is_custom_object == True
            obj = sobject.split('__')[0]
            res2 = tooling.query(f"Select Id, DeveloperName, NamespacePrefix From CustomObject where DeveloperName = '{obj}' limit 10")
            if len(res2['records']) == 0:
                utils.raiseException("RECORDS_NOT_FOUND",f"No records found for {sobject} {namespace}")
            tableEnumOrId = res2['records'][0]['Id']
        
        ns = f" and NamespacePrefix = {namespace}" if namespace != None else f" and NamespacePrefix = null "

        res2 = tooling.query(f"SELECT NamespacePrefix, DeveloperName, TableEnumOrId,ManageableState FROM CustomField where TableEnumOrId = '{tableEnumOrId}'  {ns}")

        devNames = [r['DeveloperName'] for r in res2['records']]

        q = f"SELECT NamespacePrefix, DeveloperName, TableEnumOrId,ManageableState,Metadata FROM CustomField where DeveloperName='$$$' and TableEnumOrId='{tableEnumOrId}' {ns} limit 1"
        res4 = tooling.query_threaded(q,devNames)

        for r in res2['records']:
            r['Metadata'] = [m for m in res4 if m['DeveloperName']==r['DeveloperName']][0]
            r['type'] = r['Metadata']['Metadata']['type']
            r['formula'] = True if 'formula' in r['Metadata']['Metadata'] and r['Metadata']['Metadata']['formula']!=None else False


        formulas = [r for r in res2['records'] if r['formula']==True]
        rest = [r for r in res2['records'] if r['formula']==False]

        rest.extend(formulas)
        utils.printFormated(rest,"NamespacePrefix:DeveloperName:TableEnumOrId:type:formula")

        return rest
        for r in res2['records']:
            print(f"{r['TableEnumOrId']}  {r['NamespacePrefix']} {r['DeveloperName']}")
            res3 = tooling.query(f"SELECT NamespacePrefix, DeveloperName, TableEnumOrId,ManageableState,Metadata FROM CustomField where DeveloperName='{r['DeveloperName']}' limit 1")
            r['Metadata'] = res3['records'][0]

        return res2

    def get_field_definition2(self,sobject,fieldName,namespace=None):

        ns = f" and NamespacePrefix = {namespace}" if namespace != None else f" and NamespacePrefix = null "
        res3 = tooling.query(f"SELECT NamespacePrefix, DeveloperName, TableEnumOrId,ManageableState,Metadata FROM CustomField where DeveloperName='{fieldName}' and TableEnumOrId='{sobject}' {ns} limit 1")

        print() 
        
    def no_nulls(self,obj):
        obj2 = {}
        for key in obj.keys():
            if obj[key] != None:
                if type(obj[key]) == dict:
                    obj2[key]= self.no_nulls(obj[key])
                else:
                    obj2[key] = obj[key]
        return obj2

    def calc_customField(self,sobject,field):
        name = field['name'].split('__c')[0]
        metadata = [d for d in self.definitions if d['DeveloperName'] == name][0]['Metadata']['Metadata']

        metadata2 = self.no_nulls(metadata)

        if field['name'] == 'Data_da_ultima_oportunidade__c':
            a=1
            metadata2['summaryOperation'] = metadata2['summaryOperation'].title()
            metadata2.pop('summaryFilterItems')
        data = { "FullName": f"{sobject}.{field['name']}", 
                "Metadata": metadata2
                } 

        data_str = simplejson.dumps(data, indent=2, ensure_ascii=False)

        print(data_str)
      #  print(data)
        return data

    def customField(self,sobject,field):
        if field['dependentPicklist']:
            a=1
        if field['relationshipName'] != None:
            a=1
        if field['relationshipOrder']!=None:
            a=1
        if len(field['referenceTo'])>0:
            a=1
        data = { "FullName": f"{sobject}.{field['name']}", 
                "Metadata": { "label": field['label'], 
                              "description": "my new test field", 
                              "required": False, 
                              "externalId": field['externalId'], 
                              "type": field['type'], 
                              "length": field['length'] if 'length' in field else None,
                              "unique":field['unique']
                            } 
                }
        if data['Metadata']['type'] == 'string' : 
            data['Metadata']['type'] = 'Text'

        if data['Metadata']['type'] == 'boolean' : 
            data['Metadata']['type'] = 'Checkbox'
            data['Metadata'].pop("length")
            data['Metadata']['defaultValue'] = field['defaultValue']

        if data['Metadata']['type'] == 'double' : 
            data['Metadata']['type'] = 'Number'
            data['Metadata']['precision'] = field['precision']
        #    data['Metadata']['digits'] = field['digits']
            data['Metadata']['scale'] = field['scale']
        #    data['Metadata']['length'] = field['precision']
            data['Metadata'].pop("length")

        if data['Metadata']['type'] == 'percent' : 
            data['Metadata']['type'] = 'Percent'
            data['Metadata']['precision'] = field['precision']
         #   data['Metadata']['digits'] = field['digits']
            data['Metadata']['scale'] = field['scale']

        if data['Metadata']['type'] == 'url' : 
            data['Metadata']['type'] = 'Url'
            data['Metadata'].pop("length")

        if data['Metadata']['type'] == 'textarea' : 
            data['Metadata']['type'] = 'TextArea'
            data['Metadata'].pop("length")


        if field['calculatedFormula'] != None:
      #      data['Metadata']['type'] = 'FORMULA'
         #   data['Metadata']['scale']=field['scale']
         #   data['Metadata']['calculatedFormula'] = field['calculatedFormula'] 
         #   data['Metadata']['calculated']=field['calculated']
         #   data['Metadata']['formulaTreatNullNumberAsZero']=field['formulaTreatNullNumberAsZero']
            data['Metadata']['formula']=field['calculatedFormula']
            if "length" in data['Metadata']: data['Metadata'].pop("length")

        #    if data['Metadata']['type'] == 'Text':
        #        data['Metadata']['type'] = 'Formula(Text)'
            a=1

        
        if data['Metadata']['type'] == 'reference':
            data['Metadata']['type'] = 'Lookup'
            data['Metadata']['relationshipName'] = field['relationshipName']
            data['Metadata']['referenceTo'] = field['referenceTo'][0]
            data['Metadata']['referenceTargetField'] = field['referenceTargetField'] 
            data['Metadata']['relationshipOrder'] = field['relationshipOrder'] 
            data['Metadata'].pop("length")

        if data['Metadata']['type'] == 'picklist':
            data['Metadata']['type'] = 'Picklist'
            data['Metadata'].pop("length")
            data['Metadata']['valueSet'] = {
                'restricted': field['restrictedPicklist'],
                'valueSetDefinition': {
                    'sorted': False,
                    'value': []
                }
            }
            for value in field['picklistValues']:
                value = {
                    'fullName': value['value'],
                    'default': value['defaultValue'],
                    'label': value['label']
                }
                data['Metadata']['valueSet']['valueSetDefinition']['value'].append(value)

        if data['Metadata']['type'] == 'multipicklist' : 
            data['Metadata']['type'] = 'MultiselectPicklist'
            data['Metadata'].pop("length")

            data['Metadata']['valueSet'] = {
                'restricted': field['restrictedPicklist'],
                'valueSetDefinition': {
                    'visibleLines':10,
                    'sorted': False,
                    'value': []
                }
            }
            for value in field['picklistValues']:
                value = {
                    'fullName': value['value'],
                    'default': value['defaultValue'],
                    'label': value['label']
                }
                data['Metadata']['valueSet']['valueSetDefinition']['value'].append(value)      

        print(data)
        return data

    def set_field_permission(self,sobject,data):
        action='/services/data/v54.0/composite/'
        method =  'post'        
        data={
            "allOrNone": True,
            "compositeRequest": [
                {
                    "referenceId": "Profile",
                    "url": "/services/data/v54.0/query/?q=SELECT+Id+FROM+Profile+Where+Name='System Administrator'",
                    "method": "GET"
                },
                {
                    "referenceId": "PermissionSet",
                    "url": "/services/data/v54.0/query/?q=SELECT+Id+FROM+PermissionSet+WHERE+ProfileId='@{Profile.records[0].Id}'",
                    "method": "GET"
                },
                {
                    "referenceId": "NewFieldPermission",
                    "body": {
                        "ParentId": "@{PermissionSet.records[0].Id}",
                        "SobjectType": sobject,
                        "Field": data["FullName"],
                        "PermissionsEdit": "true",
                        "PermissionsRead": "true"
                    },
                    "url": "/services/data/v54.0/sobjects/FieldPermissions/",
                    "method": "POST"
                }
            ]
        }
        call = restClient.callAPI(action=action,method=method,data=data)
        return call






import unittest,simplejson,sys
from InCli.SFAPI import restClient,query,Sobjects,utils,jsonFile,file_csv
from collections import Counter


class Test_Order_stats(unittest.TestCase):
    dec_num = 0
    orderId = '8013O000003cE7VQAU'

    def test_order_to_file(self):
        self.orderId = '8013O000003is7dQAA'  #Mobile only
        text = 'Mobile'
      #  self.orderId = '8013O000003ZDj8QAG' #Complex
      #  self.orderId = '8013O000003jEAEQA2'  #
      #  self.orderId = '8013O000003jE8XQAU' 
      #  self.orderId = '8013O000003jE7tQAE'  #MACD Diconnect and Add
      #  text='MACD_change_plan'
      #  self.orderId = '8013O000003jE2UQAU' #portabilidade
      #  text="protabilidade"
      #  self.orderId ='8013O000003jASOQA2'
      #  text="anothe"
      #  self.orderId = '8013O000003jDxDQAU'
      #  text = 'suspend_4P'
      #  self.orderId = '8013O000003jDwFQAU'
      #  text = "suspend_Interner_TV"
        self.orderId='8013O000003jDscQAE'
        text='NOS4u_100'
        self.orderId = '8013O000003jDscQAE'
        text='NOS3_new'
      #  self.orderId = '8013O000003jDrSQAU'
      #  text = 'MACD_MSISDN_change_only'

        filename = f'Decomposition_order_{self.orderId}_{text}.txt'
        original_stdout = sys.stdout

        with open(filename, 'w') as f:
            sys.stdout = f 
            self.test_order()

        sys.stdout = original_stdout 

    def get_order_items(self,orderId):
        q = f"select fields(all) from orderitem where OrderId='{orderId}' limit 200"
        order_items = query.query(q)  
        return   order_items

    def get_products_for_order_items(self,order_items):
        product2Ids = [order_item['Product2Id'] for order_item in order_items['records'] ]
        product2Ids_str = query.IN_clause(product2Ids)

        qp = f"select fields(all) from product2 where Id in ({product2Ids_str})"
        product2s = query.query(qp)
        return product2s

    def get_fulfilment_requests(self,orderId):
        q2= f"select fields(all) from vlocity_cmt__FulfilmentRequest__c where vlocity_cmt__OrderId__c = '{self.orderId}' limit 200"
        frs = query.query(q2)
        return frs
    
    def get_frls_for_frs(self,frs):
        fr_ids = [r['Id'] for r in frs['records']]
        fr_ids_str = query.IN_clause(fr_ids)

        q3= f"select fields(all) from vlocity_cmt__FulfilmentRequestLine__c where vlocity_cmt__FulfilmentRequestID__c in ({fr_ids_str}) limit 200"
        frls = query.query(q3)

        return frls

    def get_product2_for_frls(self,frls):
        rfl_product2_ids = [r['vlocity_cmt__Product2Id__c'] for r in frls['records']]
        rfl_product2_ids_str = query.IN_clause(rfl_product2_ids)
        rfl_products_q = f"select fields(all) from product2 where Id in ({rfl_product2_ids_str})"
        rfl_products = query.query(rfl_products_q)
        return rfl_products

    def test_order(self):
        restClient.init('NOSDEV')

        order_items = self.get_order_items(self.orderId)

        print(f"Order line items {len(order_items['records'])}")

        product2s = self.get_products_for_order_items(order_items)

        
        def print_decom(decom):
            print(f"       Decomposition relationship: {decom['Name']:<80}  {decom['source_Id']}->{decom['destination_Id']}  Conditions:{decom['conditions']}  mappings:{decom['mapping_rules']}")
         #   if decom['conditions_str']!=None: print(f"               {decom['conditions_str']}")
            self.dec_num = self.dec_num + 1
            if 'next_level' in decom:
                for dec in decom['next_level']:
                    print_decom(dec)

        for x,order_item in enumerate(order_items['records']):
         #   attr_str = order_item['vlocity_cmt__AttributeSelectedValues__c']
            product = [p2 for p2 in product2s['records'] if p2['Id']==order_item['Product2Id']][0]
            prod_attr_str = product['vlocity_cmt__AttributeMetadata__c']
            attributes = simplejson.loads(prod_attr_str)
            print(f"- Line {x+1} - Product:{product['Name']}  {product['ProductCode']}  Action:{order_item['vlocity_cmt__Action__c']}")
            for attribute in attributes['records']:
                for product_attributes  in attribute['productAttributes']['records']:
                    count = len(product_attributes['values']) if 'values' in product_attributes and product_attributes['inputType']=='dropdown' else 1
                    values = [value['value'] for value in product_attributes['values']] if product_attributes['inputType'] == 'dropdown' else []
                    val = ", ".join(values)
                    print(f"       Attribute: {product_attributes['inputType']:<10} ({count:>2})  {val}")

            self.get_product_decomposition(product)
            for decom in product['decompositions']:
                print_decom(decom)
           # attr = simplejson.loads(attr_str)
        print()
        print(f"total decomposition relationships {self.dec_num}")
        print()
      #  q2= f"select fields(all) from vlocity_cmt__FulfilmentRequest__c where vlocity_cmt__OrderId__c = '{self.orderId}' limit 200"
        frs = self.get_fulfilment_requests(self.orderId)

        print()
        print(f"Fulfilment Lines {len(frs['records'])}")

        frls = self.get_frls_for_frs(frs)

        rfl_products = self.get_product2_for_frls(frls)

        frls['records'][0]['vlocity_cmt__Product2Id__c']
        for y,fr in enumerate(frs['records']):
            frl_s = [r for r in frls['records'] if r['vlocity_cmt__FulfilmentRequestID__c']==fr['Id']]
            print(f" - FR: Line {y+1} {fr['Name']}  Status:{fr['vlocity_cmt__Status__c']} ")  

            for z,frl in enumerate(frl_s):
                frl_prod = [r for r in rfl_products['records'] if r['Id']==frl['vlocity_cmt__Product2Id__c']][0]
                mu = simplejson.loads(frl['vlocity_cmt__AttributesMarkupData__c'])
                at = simplejson.loads(frl['vlocity_cmt__JSONAttribute__c'])
                print(f"   - FRL: {z+1} {frl['Name']}  Product:{frl_prod['Name']}   Action:{frl['vlocity_cmt__Action__c']}")  

                for frl_att in at.keys():
                    print(f"      {frl_att} {len(at[frl_att])}")  
                    frl_att_as = at[frl_att]
                    for frl_att_a in frl_att_as:
                        count = len(frl_att_a['attributeRunTimeInfo']['values']) if frl_att_a['attributeRunTimeInfo']['dataType'] == 'Picklist' else 1
                        values = [value['value'] for value in frl_att_a['attributeRunTimeInfo']['values']] if frl_att_a['attributeRunTimeInfo']['dataType'] == 'Picklist' else []
                        val = ", ".join(values)

                        print(f"          {frl_att_a['attributeRunTimeInfo']['dataType']:<10}  ({count:>2})     {val}")

        q4 = f"select fields(all) from vlocity_cmt__OrchestrationItem__c where vlocity_cmt__OrchestrationPlanId__c = '{frs['records'][0]['vlocity_cmt__orchestrationPlanId__c']}' limit 200"
        res4 = query.query(q4)
        OrchestrationItemTypes=[]
        for rec4 in res4['records']:
            OrchestrationItemTypes.append(rec4['vlocity_cmt__OrchestrationItemType__c'])

        print()
        print()

        c = Counter(OrchestrationItemTypes)
        print(c)
        print()

    def test_objects(self):
        restClient.init('NOSQSM')
        
        res = Sobjects.get_with_only_id('a3m3O000000KCjCQAW')

        print()     
    def test_limits(self):
        restClient.init('DEVNOSCAT2')
        action = '/services/data/v51.0/limits'
        res = restClient.callAPI(action)

        print()

    def test_id(self):
        restClient.init('DEVNOSCAT2')
        action = '/services/data/v51.0'
        res = restClient.callAPI(action)

        print()

    def test_id(self):
        restClient.init('DEVNOSCAT2')
        action = '/services/data/v51.0'
        res = restClient.callAPI(action)
        for key in res.keys():
            print()
            action = res[key]
            res1 = restClient.callAPI(action)
            print(action)
            print(res1)

        print()

    def getAttributes(self,product):
        attr_str = product['vlocity_cmt__AttributeMetadata__c']
        if attr_str == None: return None
        atributes = simplejson.loads(attr_str)

        atts = []
        for atribute in atributes['records']:
            for productAttribute in atribute['productAttributes']['records']:
                if 'values' not in productAttribute:
                    a=1
                if productAttribute['values'] == None:
                    a=1
                att = {
                    'att':atribute['Code__c'],
                    'pAtt':productAttribute['code'],
                    'type':productAttribute['inputType'],
                    'len':len(productAttribute['values']) if 'values' in productAttribute and productAttribute['values'] != None else 0
                }
                atts.append(att)
        return atts

    def getChildProducts(self,product):
        children = []
        childItems = query.queryRecords(f"select fields(all) from vlocity_cmt__ProductChildItem__c where vlocity_cmt__ParentProductId__c='{product['Id']}' and vlocity_cmt__IsOverride__c = False limit 200")
        if len(childItems) == 0:
            return []

        for childItem in childItems:
            if childItem['vlocity_cmt__ChildProductId__c'] == None:
                continue
            prod = Sobjects.getF('Product2',f"Id:{childItem['vlocity_cmt__ChildProductId__c']}")['records'][0]
            print(prod['Name'])

            child = {
                'Name':childItem['vlocity_cmt__ChildProductName__c'],
                'virtual':childItem['vlocity_cmt__IsVirtualItem__c'],
                'Id':childItem['vlocity_cmt__ChildProductId__c'],
                'attributes':self.getAttributes(prod),
                'children':self.getChildProducts(prod),
                'mmq':f"({childItem['vlocity_cmt__MinMaxDefaultQty__c']})".replace(' ',''),
                "child_mm":f"[{int(childItem['vlocity_cmt__MinimumChildItemQuantity__c'])},{int(childItem['vlocity_cmt__MaximumChildItemQuantity__c'])}]"
            }
            children.append(child)
            print(childItem['vlocity_cmt__ChildProductName__c'])

        return children

    code = 'PROMO_NOS_OFFER_005'

    def test_parse_product(self):
        restClient.init('NOSDEV')

        res = Sobjects.getF('Product2',"ProductCode:C_NOS_OFFER_001")

        code='PROMO_WOO_FIXED_INTERNET_MOBILE_12_MONTHS_008'
        promo = Sobjects.getF('vlocity_cmt__Promotion__c',f"vlocity_cmt__Code__c:{self.code}")

        promoItems = query.queryRecords(f"select fields(all) from vlocity_cmt__PromotionItem__c where vlocity_cmt__PromotionId__c='{promo['records'][0]['Id']}' limit 200")
        root = {
            'children':[],
            'Name':'root',
            'Id':'NA',
            'attributes':"NA"
        }
        for promoItem in promoItems:
            prods = Sobjects.getF('Product2',f"Id:{promoItem['vlocity_cmt__ProductId__c']}")

            for prod in prods['records']:
                _product = {
                    'Name':prod['Name'],
                    'virtual':False,
                    'Id':prod['Id'],
                    'attributes':self.getAttributes(prod),
                    'children':self.getChildProducts(prod),
                    'mmq':"",
                    'child_mm':""
                }
                root['children'].append(_product)
        jsonFile.write(f'prod123_{self.code}',root)

        self.printProduct(root)

    def test_print_from_file(self):
        root = jsonFile.read(f'prod123_{self.code}')
        self.printProduct(root)

    def parse_conditions(self,condition_data):
        if condition_data == None: 
            return None

        def parse_single_condition(single_condition_a,op):
            conditions =[]

            for sc in single_condition_a:
                if 'singleconditions' in sc:
                    conditions.append(parse_single_condition(sc['singleconditions'],sc['type']))
                else:
                    condition = f"({sc['left-side']} {sc['op']} {sc['right-side']})"
                    conditions.append(condition)
            _op = f" {op} "
            condition =F"({_op.join(conditions)})"  
            return condition     

        all_conditions=[]
        for single_conditions in condition_data['singleconditions']:
            if 'singleconditions' in single_conditions:
                condition = parse_single_condition(single_conditions['singleconditions'],condition_data['type'])
            else:
                condition = f"({single_conditions['left-side']} {single_conditions['op']} {single_conditions['right-side']})"
            all_conditions.append(condition)

        if len(all_conditions)>1:
            op = F" {condition_data['type']} "
            final_condition = F"{op.join(all_conditions)}"
            return final_condition
        else:
            return all_conditions[0]
        return None

    def parse_decomposition_mapping(self,mapping,mappings):
        if mapping['mapping_type'] == 'ad-verbatim':
            if mapping['source_type'] == 'Attribute':
                mappings.append({
                    'from':mapping['source_attr_code'],
                    'to':mapping['destination_attr_code']
                })
            elif mapping['source_type'] == 'Field':
                mappings.append({
                    'from':mapping['source_field_name'],
                    'to':mapping['destination_attr_code']
                })
            else:
                a=1
        elif mapping['mapping_type'] == 'static':
            mappings.append({
                'from':'Static',
                'to':mapping['destination_attr_code']
            })

        elif mapping['mapping_type'] == 'list':
            if mapping['source_type'] == 'Attribute':
                mappings.append({
                    'from':mapping['source_attr_code'],
                    'to':mapping['destination_attr_code']
                })
            elif mapping['source_type'] == 'Field':
                mappings.append({
                    'from':mapping['source_field_name'],
                    'to':mapping['destination_attr_code']
                })
            else:
                a=1
        else:
            a=1
 
    def get_product_decomposition(self,product):
        q = f"select fields(all) from vlocity_cmt__DecompositionRelationship__c where vlocity_cmt__SourceProductId__c='{product['Id']}' limit 200"
        decomposition_relationships = query.query(q)
        product['decompositions'] = []
        for decomposition_relationship in decomposition_relationships['records']:
            mappring_data_str = decomposition_relationship['vlocity_cmt__MappingsData__c']
            mappring_data = simplejson.loads(decomposition_relationship['vlocity_cmt__MappingsData__c']) if mappring_data_str != None else None

            mappings =[]
            if mappring_data != None:
                for mapping in mappring_data:
                    self.parse_decomposition_mapping(mapping,mappings)

            condition_data_str = decomposition_relationship['vlocity_cmt__ConditionData__c']
            condition_data = simplejson.loads(condition_data_str) if condition_data_str!=None else None

            self.parse_conditions(condition_data)

            decomposition = {
                'conditions' :len(condition_data['singleconditions']) if condition_data != None else 0,
                'conditions_str':self.parse_conditions(condition_data),
                'mapping_rules':len(mappring_data) if mappring_data != None else 0,
                'Name':decomposition_relationship['Name'],
                'destination_Id':decomposition_relationship['vlocity_cmt__DestinationProductId__c'],
                'source_Id':decomposition_relationship['vlocity_cmt__SourceProductId__c'],
                'mappings':mappings
            }
            product['decompositions'].append(decomposition)

            if decomposition['destination_Id'] != None:
                q = f"select fields(all) from vlocity_cmt__DecompositionRelationship__c where vlocity_cmt__SourceProductId__c='{decomposition['destination_Id']}' limit 200"
                res2 = query.query(q)
                if len(res2['records'])>0:
                    level = 2 if 'level' not in product else product['level']+1
                    fake_product = {
                        'Id':decomposition['destination_Id'],
                        'level':level
                    }
                    self.get_product_decomposition(fake_product)
                    decomposition['next_level'] = fake_product['decompositions']
            else:
                a=1

    def test_decomposition_rules(self):
        restClient.init('NOSDEV')

        root = jsonFile.read(f'prod123_{self.code}')

        def get_decomposition(product):
            self.get_product_decomposition(product)
            for child in product['children']:
                get_decomposition(child)

        for children in root['children']:
            get_decomposition(children)

        jsonFile.write(f'prod123_decomposed_{self.code}',root)
        print()

    def flatten(self,product):
        a=1

    def printProduct(self,products,path=[]):

        filename = f'prod123_decomposed_{self.code}_csv.csv'
        original_stdout = sys.stdout

        with open(filename, 'w') as f:
            sys.stdout = f 

            def printProduct_inner(products,path=[]):
                if products == None:
                    a=1
                for product in products['children']:
                    spath = path.copy()
                    spath.append(f"{product['Name']}{product['mmq']}{product['child_mm']}")

                    self.printAttribute(spath,product['attributes'])
                    self.print_decomposition(spath,product['decompositions'])
                    try:
                        printProduct_inner(product,spath)
                    except Exception as e:
                        print(e)
            printProduct_inner(products)
        sys.stdout = original_stdout 

        print()


    def printAttribute(self,path,attributes):
        spath = path.copy()
        while len(spath)<5:
            spath.append("")

        _path = ";".join(spath)

        if attributes == None:  return

        for atttribute in attributes:
            print(f"{_path};{atttribute['att']}-{atttribute['pAtt']};{atttribute['type']};{atttribute['len']}")

    def print_decomposition(self,path,decompostions):
        spath = path.copy()
        while len(spath)<8:
            spath.append("")

        _path = ";".join(spath)

        if decompostions == None:      return

        def print_next_level(path,next_decompositions,level):
            for next_decomposition in next_decompositions:
                xx = []
                while len(xx)< (3 + (level-1)):
                    xx.append("")
                spaces = ";".join(xx)
                for next_mapping in next_decomposition['mappings']:
                    print(f"{path};{spaces};{next_decomposition['Name']};{next_mapping['from']}->{next_mapping['to']}")
                if 'next_level' in next_decomposition:
                    print_next_level(f"{_path};{decomposition['Name']};{next_decomposition['Name']}",next_decomposition['next_level'],level+1)

        for decomposition in decompostions:
            for mapping in decomposition['mappings']:
                print(f"{_path};{decomposition['Name']};{decomposition['conditions']};{decomposition['mapping_rules']};{mapping['from']}->{mapping['to']}")
            if 'next_level' in decomposition:
                print_next_level(f"{_path};{decomposition['Name']}",decomposition['next_level'],1)

    def test_print_from_file_deco(self):
        root = jsonFile.read(f'prod123_decomposed_{self.code}')
        self.printProduct(root)

#####################################################################
    def printProduct2(self,products,path=[]):

        filename = f'prod123_decomposed2_{self.code}_csv.csv'
        original_stdout = sys.stdout

        with open(filename, 'w') as f:
            sys.stdout = f 

            def printProduct_inner(products,path=[]):
                if products == None:
                    a=1
                for product in products['children']:
                    spath = path.copy()
                    virtual = " VIRTUAL " if product['virtual'] == True else ""
                    spath.append(f"{virtual}{product['Name']}{product['mmq']}{product['child_mm']}")

                    self.printAttribute2(spath,product['attributes'])
                    self.print_decomposition2(spath,product['decompositions'])
                    try:
                        printProduct_inner(product,spath)
                    except Exception as e:
                        print(e)
            printProduct_inner(products)
        sys.stdout = original_stdout 

        print()


    def printAttribute2(self,path,attributes):
        spath = path.copy()
        while len(spath)<5:
            spath.append("")

        _path = ";".join(spath)

        if attributes == None: return
        for atttribute in attributes:
            att = f"{atttribute['att']}-{atttribute['pAtt']}"
            print(f"{_path};AT:  {att};AT:{atttribute['type']} {atttribute['len']}")


    def print_decomposition2(self,path,decompostions):
        spath = path.copy()
        while len(spath)<5:
            spath.append("")

        _path = ";".join(spath)

        if decompostions == None: 
     #       print(f"{_path}")
            return

        def print_next_level(path,next_decompositions,level):
            for next_decomposition in next_decompositions:
                xx = []
                while len(xx)< (3 + (level-1)):
                    xx.append("")
                spaces = ";".join(xx)
                spaces = ''
                for next_mapping in next_decomposition['mappings']:
                    print(f"{path};DE: {next_decomposition['Name']};MAP: {next_mapping['from']}->{next_mapping['to']}")
                if 'next_level' in next_decomposition:
                    print_next_level(f"{_path};DE: {decomposition['Name']};DE: {next_decomposition['Name']}",next_decomposition['next_level'],level+1)

        for decomposition in decompostions:
            for mapping in decomposition['mappings']:
                print(f"{_path};DE: {decomposition['Name']}  C:{decomposition['conditions']};MAP: {mapping['from']}->{mapping['to']}")
            if 'next_level' in decomposition:
                print_next_level(f"{_path};DE: {decomposition['Name']}  C:{decomposition['conditions']}",decomposition['next_level'],1)

    def test_print_from_file_deco2(self):
        root = jsonFile.read(f'prod123_decomposed_{self.code}')
        self.printProduct2(root)


import unittest,simplejson
from InCli import InCli
from InCli.SFAPI import file,DR_IP,restClient,query,utils,thread,jsonFile,file_csv

class Test_DR_IP(unittest.TestCase):
    options = {
        "isDebug": True,
        "chainable": True,
        "resetCache": False,
        "ignoreCache": True,
        "queueableChainable": False,
        "useQueueableApexRemoting": False
    }     
    def test_IP(self):
        #restClient.init('DTI')
        restClient.init('NOSDEV')

        
        call = DR_IP.ip("custom_GetTrialPromos",input={},options=self.options)
        lc = restClient.lastCall()

        print()

    def test_IP_test(self):
        restClient.init('DTI')

        input = {
            "cartId": "8010Q0000035ec1QAA"
        }
        
        call = DR_IP.ip("unai_chainableIpsTest",input=input,options=self.options)


        print()

    key ='d18120d8-9217-43f9-72b5-9fce1b3cdcd8'
    def test_attachment(self):
        restClient.init('DTI')

        q = f"select fields(all) from vlocity_cmt__DRBulkData__c where vlocity_cmt__GlobalKey__c = '{self.key}' limit 20"
        call0 = query.query(q)
        print(call0)
        
        q2 = f"select fields(all) from Attachment where ParentId ='{call0['records'][0]['Id']}' limit 10"
        call2 = query.query(q2)
        print(call2['records'][0]['Id'])
        attachmentId = call2['records'][0]['Id']
     #   attachmentId = '00P0Q00000JWEzGUAX'
        action = call2['records'][0]['Body']
        call = restClient.requestWithConnection(action=action)

        filepath = restClient.callSave("AttachementX123")

        print()

    def test_finish_call(self):
        restClient.init('DTI')
        input = "{}"

        options1 = self.options.copy()
        options1['vlcIPData'] = self.key

        call = DR_IP.ip("unai_chainableIpsTest",input=input,options=options1)

        print()

    def test_dr_bundle(self):
        restClient.init('DTI')

        q = "SELECT Name, Id, LastModifiedDate, LastModifiedBy.Name, CreatedDate, CreatedBy.Name, vlocity_cmt__Type__c, vlocity_cmt__InputType__c, vlocity_cmt__OutputType__c, vlocity_cmt__Description__c, LastModifiedById FROM vlocity_cmt__DRBundle__c USING SCOPE Everything WHERE vlocity_cmt__Type__c != 'Migration' AND  vlocity_cmt__Type__c != 'Export (Component)' ORDER BY Name"

        q = "SELECT Name, Id, LastModifiedDate, LastModifiedBy.Name, CreatedDate, CreatedBy.Name, vlocity_cmt__Type__c, vlocity_cmt__InputType__c, vlocity_cmt__OutputType__c, vlocity_cmt__Description__c, LastModifiedById FROM vlocity_cmt__DRBundle__c ORDER BY Name"

        res = query.query(q)
        
        out = []
        for record in res['records']:
            o = {
                "Name":record['Name'],
                "type":record['vlocity_cmt__Type__c']

            }
            out.append(o)
        utils.printFormated(out)
        print()

    def bestMach(self,log,ip_definitions):
        possible_ips = []

        sequenceExecute = log['vlcDebug']['executionSequence']
        sequence = [key[0:-6] for key in log.keys() if key.endswith('Status') and type(log[key]) is bool]
     #   for key in sequence:
     #       if key not in log:
     #           sequence.remove(key)

        for ip_definition in ip_definitions:
            if ip_definition['vlocity_cmt__ProcedureKey__c'] == 'iUse_submitResumeOrder':
                a=1
            score = 0

            executed = ""
            for ip_step in ip_definition['steps']:
                ex = "1" if ip_step['name'] in sequence else "0"
                if ex == "0" and ip_step['executionConditionalFormula'] != "": ex = 'C'
                if ex == "0" and ip_step['loop_element'] == True: ex = 'L' 
                executed = f"{executed}{ex}"

            missed = []
            for seq in sequence:
                if seq not in ip_definition['steps_names']:
                    executed = f"{executed}*"
                    missed.append(seq)
            score = executed.count('1')

            if score>0:
                ip = {
                    'name':ip_definition['vlocity_cmt__ProcedureKey__c'],
                    'score':score,
                    'ip_steps':len(ip_definition['steps']),
                    'size':len(sequence),
                    'execured':executed,
                    'ip_stesps':ip_definition['steps'],
                    'debug_steps':sequence,
                    'missed':missed
                }
                possible_ips.append(ip)
        
        possible_ips_sorted = sorted(possible_ips, key=lambda d: d['score'])
        possible_ips_sorted.reverse()

        return possible_ips_sorted


    def findMatch(self,log,ip_definitions):
        possible_ips = []

        sequenceExecute = log['vlcDebug']['executionSequence']
        sequence = [key[0:-6] for key in log.keys() if key.endswith('Status') and type(log[key]) is bool]
        if sorted(sequence) == sorted(sequenceExecute):
            a=1
        else:
            a=1

        for ip_definition in ip_definitions:           
            if len(sequence) > len(ip_definition['steps_names']):  continue

            if ip_definition['vlocity_cmt__ProcedureKey__c'] == 'woo_getBundleNBOOffer':
                a=1          
            found = True
            for step in sequence:
                if step not in ip_definition['steps_names']: 
                    found = False
                    break
            if found: 
                posible = {
                    'ip':ip_definition['vlocity_cmt__ProcedureKey__c'],
                    'executed':"",
                    'ip_steps':ip_definition['steps_names'],
                    'debug_steps':sequence,
                    'missing':[]
                }
                for ip_step in ip_definition['steps']:
                    ex = "1" if ip_step['name'] in sequence or ip_step['name'] in sequenceExecute else "0"
                    if '0' not in posible['executed']:
                        if ex == "0" and ip_step['executionConditionalFormula'] != "": 
                            ex = 'C'
                        if ex == "0" and ip_step['loop_element'] == True: ex = 'L'
                    if ex == '0': posible['missing'].append(ip_step)

                    posible['executed'] = f"{posible['executed']}{ex}"
                if '1' not in posible['executed']: continue
                if posible['executed'][0]=='0': 
                    continue
                if '01' in posible['executed']:
                    continue
                possible_ips.append(posible)

        return possible_ips

    def test_find_stooped_VIPs_threaded(self):
        restClient.init('NOSPRD')


        if 1==2:
            q0 = f"select Id,vlocity_cmt__GlobalKey__c from  vlocity_cmt__DRBulkData__c where vlocity_cmt__DRBundleName__c = 'None Specified' and vlocity_cmt__AsyncApexJobId__c = null and vlocity_cmt__GlobalKey__c != null order by Id desc limit 200"
            bulk_data_records = query.query(q0)

            bulk_ids = [record['Id'] for record in bulk_data_records['records']]

            q = f"select ID,Body,ParentId,LastModifiedDate from Attachment where ParentId in ({query.IN_clause(bulk_ids)})"
            q = f"select fields(all) from Attachment where ParentId in ({query.IN_clause(bulk_ids)}) limit 200"

            attachments = query.query_base64(q)

            for attachment in attachments['records']:
                attachment['vlocity_cmt__GlobalKey__c'] = [bdr['vlocity_cmt__GlobalKey__c'] for bdr in bulk_data_records['records'] if bdr['Id']==attachment['ParentId']][0]

            jsonFile.write('attachments_123',attachments)
            
        else:
            attachments = jsonFile.read('attachments_123')

        self.process_attachements_threaded(attachments)

    def test_process_one(self):
        restClient.init('DTI')

        bulkDataId = 'a2J0Q000001Qim1UAC'
        q0 = f"select Id,vlocity_cmt__GlobalKey__c from  vlocity_cmt__DRBulkData__c where Id='{bulkDataId}'"
        bulk_data_records = query.query(q0)

        q = f"select fields(all) from Attachment where ParentId = '{bulkDataId}' limit 1"
        attachments = query.query_base64(q)

        for attachment in attachments['records']:
            attachment['vlocity_cmt__GlobalKey__c'] = [bdr['vlocity_cmt__GlobalKey__c'] for bdr in bulk_data_records['records'] if bdr['Id']==attachment['ParentId']][0]

        self.process_attachements_threaded(attachments)


    def test_print_attachments(self):
        restClient.init('NOSQSM')
        date1 = "2023-02-24"


        q0 = f"select Id,vlocity_cmt__GlobalKey__c from  vlocity_cmt__DRBulkData__c where vlocity_cmt__DRBundleName__c = 'None Specified' and vlocity_cmt__AsyncApexJobId__c = null and vlocity_cmt__GlobalKey__c != null and LastModifiedDate>{date1}T00:00:00.00Z "
        bulk_data_records = query.query(q0)

        q = f"select ID,Body,ParentId,LastModifiedDate from Attachment where ParentId in (select Id from  vlocity_cmt__DRBulkData__c where vlocity_cmt__DRBundleName__c = 'None Specified' and vlocity_cmt__AsyncApexJobId__c = null and vlocity_cmt__GlobalKey__c != null and LastModifiedDate>{date1}T00:00:00.00Z)"

        attachments = query.query_base64(q)   

        for attachment in attachments['records']:
            attachment['vlocity_cmt__GlobalKey__c'] = [bdr['vlocity_cmt__GlobalKey__c'] for bdr in bulk_data_records['records'] if bdr['Id']==attachment['ParentId']][0]
            attachment['log'] = restClient.requestWithConnection(action=attachment['Body'])
            if 'UpdateOrderStatusToPaid_RA' in attachment['log']:
       #     if 'GetOrderServiceAccount_DEA' in attachment['log']:

                print(attachment['log']['OrderDetailsNode']['OrderNumber'])
                json_formatted_str = simplejson.dumps(attachment['log'], indent=2, ensure_ascii=False)

                print(json_formatted_str)
        print()

    def test_date(self):
        restClient.init('NOSPRD')
        date1 = "2023-01-01"
        date2 = "2023-02-01"


        q0 = f"select Id,vlocity_cmt__GlobalKey__c from  vlocity_cmt__DRBulkData__c where vlocity_cmt__DRBundleName__c = 'None Specified' and vlocity_cmt__AsyncApexJobId__c = null and vlocity_cmt__GlobalKey__c != null and LastModifiedDate>{date1}T00:00:00.00Z and LastModifiedDate<{date2}T00:00:00.00Z"
        bulk_data_records = query.query(q0)

        q = f"select ID,Body,ParentId,LastModifiedDate from Attachment where ParentId in (select Id from  vlocity_cmt__DRBulkData__c where vlocity_cmt__DRBundleName__c = 'None Specified' and vlocity_cmt__AsyncApexJobId__c = null and vlocity_cmt__GlobalKey__c != null and LastModifiedDate>{date1}T00:00:00.00Z and LastModifiedDate<{date2}T00:00:00.00Z)"

        attachments = query.query_base64(q)

        for attachment in attachments['records']:
            attachment['vlocity_cmt__GlobalKey__c'] = [bdr['vlocity_cmt__GlobalKey__c'] for bdr in bulk_data_records['records'] if bdr['Id']==attachment['ParentId']][0]

        self.process_attachements_threaded(attachments)


        print()



    num_threads=15
    def process_attachements_threaded(self,attachments):
        ip_definitions = self.get_IP_definitions()

        result = []

        def do_work(attachment):
            attachment['log'] = restClient.requestWithConnection(action=attachment['Body'])
            return attachment

        def on_done(attachment,result):
            log = attachment['log']
            if 'vlcDebug' in log:
                possible_ips =self.findMatch(log,ip_definitions)
                if len(possible_ips) >0:  attachment['possible'] = possible_ips
                else: attachment['best_match'] = self.bestMach(log,ip_definitions)
                result.append(attachment)

        thread.execute_threaded(attachments['records'],result,do_work,on_done,threads=self.num_threads)

        found = [attachment for attachment in attachments['records'] if 'possible' in attachment]
        best_match = [attachment for attachment in attachments['records'] if 'best_match' in attachment]

        newlist = sorted(found, key=lambda d: d['LastModifiedDate']) 

        print_list = []
        for record in newlist:
            for pos in record['possible']:
                p = {
                    'LastModifiedDate':record['LastModifiedDate'][0:19],
                    'Id':record['Id'],
                    'ParentId':record['ParentId'],
                    'vlocity_cmt__GlobalKey__c':record['vlocity_cmt__GlobalKey__c'],
                    'IP':pos['ip'],
                    'debug':pos['executed'],
                    'missing':",".join([f"{rec['name']}-[{rec['type']}]" for rec in pos['missing']])
                }
                print_list.append(p)


        file_csv.write('Attachment_records_Jan',print_list)
        utils.printFormated(print_list)

        print()


    def test_find_stooped_VIPs(self):
        restClient.init('NOSQSM')

        q0 = f"select Id,vlocity_cmt__GlobalKey__c from  vlocity_cmt__DRBulkData__c where vlocity_cmt__DRBundleName__c = 'None Specified' and vlocity_cmt__AsyncApexJobId__c = null and vlocity_cmt__GlobalKey__c != null"
        bulk_data_records = query.query(q0)
        bdrs={}
        for bdr in bulk_data_records['records']: bdrs[bdr['Id']] = bdr['vlocity_cmt__GlobalKey__c']
        print(len(bdrs))

        q = f"select ID,Body,ParentId from Attachment where ParentId in (select Id from  vlocity_cmt__DRBulkData__c where vlocity_cmt__DRBundleName__c = 'None Specified' and vlocity_cmt__AsyncApexJobId__c = null and vlocity_cmt__GlobalKey__c != null)"

        attachments = query.query_base64(q)

        for attachment in attachments['records']:
            attachment['vlocity_cmt__GlobalKey__c'] = bdrs[attachment['ParentId']]

        ip_definitions = self.get_IP_definitions()

        ip_definitions_woo = [ip_definition for ip_definition in ip_definitions if ip_definition['vlocity_cmt__ProcedureKey__c'].startswith('woo_')]

        for ip_definition_woo in ip_definitions_woo:
            ip_definition_woo['steps'] = set(ip_definition_woo['steps'])
        for attachment in attachments['records']:
            log = restClient.requestWithConnection(action=attachment['Body'])
            possible_ips = []

            if 'vlcDebug' in log:
                sequence = log['vlcDebug']['executionSequence']
                possible_ips =self.findMatch(sequence,ip_definitions_woo)
                if len(possible_ips) >0:
                    bulk_data_record = [record for record in bulk_data_records['records'] if record['Id']==attachment['ParentId']][0].copy()
                    bulk_data_record['possible'] = possible_ips
                    utils.printFormated(bulk_data_record)
            #    else:
            #        posibles = self.bestMach(sequence,ip_definitions)
           #         utils.printFormated(posibles)

    def get_IP_definitions(self):
        q = f"""select 
                    Id,
                    vlocity_cmt__Content__c,
                    vlocity_cmt__OmniScriptId__r.name,
                    vlocity_cmt__OmniScriptId__r.vlocity_cmt__ProcedureKey__c 
                    from vlocity_cmt__OmniScriptDefinition__c 
                    where vlocity_cmt__OmniScriptId__c in (select Id from vlocity_cmt__OmniScript__c where vlocity_cmt__OmniProcessType__c = 'Integration Procedure' and vlocity_cmt__IsActive__c = TRUE) """

        res = query.query(q)

        ip_definitions = []

        for record in res['records']:
            ip_definition = {
                'vlocity_cmt__ProcedureKey__c': record['vlocity_cmt__OmniScriptId__r']['vlocity_cmt__ProcedureKey__c'],
                'Name':                         record['vlocity_cmt__OmniScriptId__r']['Name'],
                'steps':                        [],
                'steps_names':[]
            }
            if ip_definition['vlocity_cmt__ProcedureKey__c'] == 'MACD_ichangeOrderProcessAsynch':
                a=1
            ip_definitions.append(ip_definition)
            content = simplejson.loads(record['vlocity_cmt__Content__c'])

            def getChildStep(child,parent_conditional='',parent_loop=False):
                if 'installCoverageEligibility' == child['name']:
                    a=1
                step = {
                    'name':child['name'],
                    'label':child['propSetMap']['label'] if 'label' in child['propSetMap'] else '',
                    'executionConditionalFormula':child['propSetMap']['executionConditionalFormula'] if 'executionConditionalFormula' in child['propSetMap'] else '',
                    'chainOnStep':child['propSetMap']['chainOnStep'] if 'chainOnStep' in child['propSetMap'] else False,
                    'indexInParent':child['indexInParent'],
                    'type':child['type'],
                    'loop_element':parent_loop
                }    
                if parent_conditional != '':
                    step['executionConditionalFormula'] = f"{step['executionConditionalFormula']} - {parent_conditional}"

                if child['type'] == 'Loop Block':
                    step['loop_element'] = True

                return step

            def addChildren(child,parent_conditional='',parent_loop=False):
                step = getChildStep(child,parent_conditional,parent_loop)
                ip_definition['steps'].append(step)
                ip_definition['steps_names'].append(step['name'])

                if 'children' in child and len(child['children'])>0:
                    for child1 in child['children']:
                        if 'eleArray' in child1:
                            for eleChild in child1['eleArray']:
                                parent_loop = True if child['type'] == 'Loop Block' or parent_loop == True  else False
                                addChildren(eleChild,parent_conditional=step['executionConditionalFormula'],parent_loop=parent_loop)
                        else:
                            a=1

            for child in content['children']:
                addChildren(child)

        return ip_definitions



import logging, copy
from . import Sobjects, objectUtil, priceBook, query as queryM, utils,restClient
from datetime import date
import simplejson 


_carts = []
def _get(Id):
    for cart in _carts:
        if cart['Id'] == Id:
            return cart
def _set(Id,name,pricelistId,accountId):
    global _carts
    _carts.append({
        'Id':Id,
        'name':name,
        'plId':pricelistId,
        'accId':accountId
    })
#------------------------------------------------------------------------

def stringify(parameters,exclude,initSeparator='?'):
  isFirst = True

  ret = ''
  for name in parameters:
    if name in exclude:
      continue
    if(parameters[name] == None):
      continue
    separator = initSeparator if isFirst == True else '&'
    value = parameters[name]
    if isinstance(value, str) == False:
      value = simplejson.dumps(value)
    paramStr = f'{separator}{name}={value}'
    ret = ret + paramStr
    isFirst = False

  return ret


#------------------------------------------------------------------------
def checkError():
    lc = restClient.lastCall()
    if lc['error'] is not None:
      utils.raiseException(lc['errorCode'],lc['error'],lc['action'])
    response = lc['response']
    if 'messages' in response:
        if len(response['messages']) > 0:
            msg = response['messages'][0]
            if 'code' in msg:
                if msg['code'] == '208' or msg['code'] == '101': return False

                if msg['severity'] == 'ERROR':  utils.raiseException(msg['code'],msg['message'])

    return False

#------------------------------------------------------------------------
def getCartId(orderF):
    cartId = Sobjects.IdF('Order',orderF)
    return cartId

def createQuote(opportunityid,pricelistName,name, recordTypeId,expirationDate = "2023-01-17" ,inputFields = None,fields='Id,Name,EffectiveDate'):
    global _carts

    Body = {
    "methodName": "createCart",
    "objectType": 'Quote',
    "inputFields": [
        {"opportunityid": opportunityid},
        {"Name": name},
        {"ExpirationDate": expirationDate},
        {"RecordTypeId": recordTypeId},
        {"pricelistName": pricelistName}
        ],
    "subaction":'createQuote',
    "fields": fields
    }     

    if inputFields != None:
        Body['inputFields'] = Body['inputFields'] + inputFields

    call = restClient.callAPI('/services/apexrest/vlocity_cmt/v2/carts',
                                method="post",
                                data=Body)
    logging.info(f"Quote {name} created with code {restClient.lastCall()['status_code']}")

  #  _set(call['records'][0]['Id'],name,pricelistId,accountId)
    checkError()
    return  call

def createCart_api(accountId,pricelistId,name='', subaction='createOrder',effectivedate =  date.today().strftime("%Y-%m-%d"),channel='python',inputFields = None,fields='Id,Name,EffectiveDate'):
    """
    - subaction: createOrder or createQuote
    RETURNS: ['records'][0]['Id']"""
    global _carts

    objectType = "Order"
    if subaction == 'createQuote':
        objectType = "Quote"

    if subaction == 'createOrder':
        Body = {
            "methodName": "createCart",
            "objectType": objectType,
            "inputFields": [
                {"effectivedate": effectivedate },
                {"status": "Draft"},
                {"Name": name},
                {"AccountId": accountId},
                {"vlocity_cmt__PriceListId__c": pricelistId}, 
                {"vlocity_cmt__OriginatingChannel__c": channel}
                ],
            "subaction":subaction,
            "fields": fields
            }

    if inputFields != None:
        Body['inputFields'].append(inputFields)

    call = restClient.callAPI('/services/apexrest/vlocity_cmt/v2/carts',
                                method="post",
                                data=Body)

    checkError()
    restClient.glog().info(f"Cart {name} created with Id {call['records'][0]['Id']} for {accountId}")

   # _set(call['records'][0]['Id'],name,pricelistId,accountId)

    return call

def createCart(accountF,pricelistF,name,checkExists=False,subaction='createOrder',effectivedate =  date.today().strftime("%Y-%m-%d"),channel='python',inputFields=None,fields='Id,Name,EffectiveDate'):
    """
    As per create cart API.
    - accountF as extended field
    - pricelistF as extended field
    - name: the name for the cart
    - checkExists: will query for a cart with that name. if exist will return it and not create another one. 
    - RETURNS: the cart Id
    """
    if type(accountF) is dict:
        accountId = utils.Id(accountF)
    else:
        accountId = queryM.queryIdF('Account',accountF)
    
    pricelistId = queryM.queryIdF('vlocity_cmt__PriceList__c',pricelistF)

    if checkExists:
        cartId = queryM.queryField(f" select Id from Order where Name ='{name}' and AccountId='{accountId}' ")
        if cartId != None:
            restClient.glog().info(f"Cart with Name {name} exists. Returning existing")
            _set(cartId,name,pricelistId,accountId)
            return cartId

    cartCall= createCart_api(accountId,pricelistId,name,subaction,effectivedate,channel,inputFields,fields)
    cartId = cartCall['records'][0]['Id']
    return cartId

def createCartFromAsset(assetId, accountId, date, inputFields=None):
    """As per assetToOder API
    - inputFields: sets the fields in the Cart. Json of the form {"Name":"value"}
    - return value is the cart Id, or an exception is thrown
    """
    call = createCartFromAsset_api(assetId,accountId,date)

    id = call['records'][0]['cartId']

    if inputFields != None:
        data = inputFields
        Sobjects.update(id, data)
        restClient.glog().info(f"Cart updated with inputFields {inputFields}.")

    return id

def createCartFromAsset_api(assetId, accountId, date):
    """As per the API
    - return ['records'][0]['cartId']"""
    if type(assetId) == list:
        assetId = ",".join(assetId)

    data = {
        "subaction": "assetToOrder",
        "id": assetId,
        "accountId": accountId,
        "requestDate": date
    }

    call = restClient.callAPI('/services/apexrest/vlocity_cmt/v2/carts',method="post",data=data)
    checkError()
    restClient.glog().info(f"Created Cart {call['records'][0]['cartId']} from asset {assetId} for account {accountId}")

    return call

def getCartSummary_api(cartid,validate=None,price=None,headerFieldSet=None,translation=None):

    paramStr = stringify(locals(),exclude=['cartid'])
    action = f'/services/apexrest/vlocity_cmt/v2/cpq/carts/{cartid}{paramStr}'

    call = restClient.callAPI(action)
    checkError()
    return call

def deleteCart(cartId,cartType='Order'):
    delete = Sobjects.delete('Order',cartId)
    checkError()
    restClient.glog().info(f"Cart {cartId} deleted.")
    return delete

def get_item_attributes_NO(cart_ID,item_ID):
    action = f'/services/apexrest/vlocity_cmt/v2/cpq/carts/{cart_ID}/items/{item_ID}/itemAttributes'

    call = restClient.callAPI(action)
    checkError()
    return call

def update_item_attribute_api(cartId,item):
    itemId = Sobjects.Id(cartId)
   # itemId = item['Id']['value']
    action = f'/services/apexrest/vlocity_cmt/v2/cpq/carts/{cartId}/items/{itemId}/itemAttributes'
    method ='Put'
    params= {
        "methodName": "putItemAttributes",
        "items": {
            "records": [
                item
            ]
        },
        "filters": None,
        "itemId": itemId,
        "id": itemId,
        "cartId": cartId,
        "price":False,
        "validate":False
    }
    call = restClient.callAPI(action,method=method,data=params)
    filepath = restClient.callSave('attributes123',logRequest=True,logReply=False)

    checkError()
    return call

def runCartValidation_api(cartid,validate=None,price=None):

    paramStr = stringify(locals(),exclude=['cartid'])
    action = f'/services/apexrest/vlocity_cmt/v2/cpq/carts/{cartid}{paramStr}'

    body=    {
        "cartId":cartid,
        "methodName":"runCartValidation"#,
      #  "price":true,
      #  "validate":true
        }
    call = restClient.callAPI(action,method='post',data=body)
    checkError()
    return call

def getCartItems_api(cartId, query=None,id=None,lastRecordId=None,pagesize=None,hierarchy=None,includeAttachment=None,headerFieldSet=None,filter=None,price=None,validate=None,fields=None):
    paramStr = stringify(locals(),exclude=['cartId'])

    action = f'/services/apexrest/vlocity_cmt/v2/cpq/carts/{cartId}/items{paramStr}'

    call = restClient.callAPI(action)
    checkError()
    return call

def addItemstoCart(cartId,productCode,pricelistF,parentId=None,parentHierarchyPath=None,parentRecord=None,price=True,validate=True,hierarchy=None,pagesize=None,includeAttachment=None,expandAll=None,fields=None,noResponseNeeded=False):
    plEF = utils.extendedField(pricelistF)
    pb2Id = queryM.queryField(f" select vlocity_cmt__Pricebook2Id__c from vlocity_cmt__PriceList__c where {plEF['name']} = '{plEF['value']}' ")
    pricebookEntryId = queryM.queryField(f" select Id from PricebookEntry where ProductCode='{productCode}' and Pricebook2Id = '{pb2Id}' ")

    priceBookEntryId = priceBook.pricebookEntryId_pl(_get(cartId)['plId'],productCode,pricelistField='Id')
    return addItemstoCart_api(cartId,priceBookEntryId,parentId,parentHierarchyPath,parentRecord,price,validate,hierarchy,pagesize,includeAttachment,expandAll,fields,noResponseNeeded)

def addItemstoCart_api(cartId,priceBookEntryId,parentId=None,parentHierarchyPath=None,parentRecord=None,price=True,validate=True,hierarchy=None,pagesize=None,includeAttachment=None,expandAll=None,fields=None,noResponseNeeded=False):
    paramStr = stringify(locals(),exclude=['cartId','itemIds','parentRecord','priceBookEntryId'])

    action = f'/services/apexrest/vlocity_cmt/v2/cpq/carts/{cartId}/items{paramStr}'

    data = {
        "items": [{
            "itemId": priceBookEntryId,  
        }]
    }

    if parentRecord != None:
        data = {
            "methodName": "postCartsItems",
            "items": [{
                "parentId": parentId,
                "parentHierarchyPath": parentHierarchyPath,
                "itemId": priceBookEntryId,  
                "parentRecord": {
                    "records": [
                        parentRecord
                    ]
                }

            }],
            "cartId": cartId,
            "price": price,
            "validate": validate,
            "includeAttachment": False,
            "pagesize": 10,
            "lastRecordId": None,
            "hierarchy": -1,
            "query": None
        }
    if noResponseNeeded == True:
        data['noResponseNeeded'] = True

    call = restClient.callAPI(action,method='post',data=data)

    restClient.glog().info(f"added {priceBookEntryId} to cart {cartId}")
    checkError()
    return call

def updateCartItem_api(cartId, items):

    action = f'/services/apexrest/vlocity_cmt/v2/cpq/carts/{cartId}/items'
    data = {
        "methodName": "putCartsItems",
        "items": items,
        "price": False,
        "validate": False,
        "includeAttachment": False ,
      #  "pagesize": 10,
      #  "lastRecordId": None,
        "hierarchy": 5#,
      #  "query": None,
      #  "filters": None,
       # "fields": "vlocity_cmt__BillingAccountId__c,vlocity_cmt__ServiceAccountId__c,Quantity,vlocity_cmt__RecurringTotal__c,vlocity_cmt__OneTimeTotal__c,vlocity_cmt__OneTimeManualDiscount__c,vlocity_cmt__RecurringManualDiscount__c,vlocity_cmt__ProvisioningStatus__c,vlocity_cmt__RecurringCharge__c,vlocity_cmt__OneTimeCharge__c,ListPrice,vlocity_cmt__ParentItemId__c,vlocity_cmt__BillingAccountId__r.Name,vlocity_cmt__ServiceAccountId__r.Name,vlocity_cmt__PremisesId__r.Name,vlocity_cmt__InCartQuantityMap__c,vlocity_cmt__EffectiveQuantity__c"
        }
    call = restClient.callAPI( action, method="put",data=data)
    checkError()
    return call

def _getItemAtPath(obj,path,field='ProductCode'):
    for p in path.split(':'):
        obj = objectUtil.getSibling(obj,field,p)  
    return obj

def addToCart_action(parentItem,product,field='ProductCode'):
    sibbling = objectUtil.getSiblingWhere(parentItem,field,product,'itemType','childProduct')['object']
    if sibbling == None:
        logging.warn("The expand does not have a reference to the product. {field}:{product}")
        return None

    actions = objectUtil.getSiblingWhere(sibbling,selectKey='actions')['object']

    if 'records' in parentItem:
        parentItem = parentItem['records'][0]
    parentItem = copy.deepcopy(parentItem)
    parentItem['childProducts'] = ""   

    call = executeActions(actions,'addtocart',parentItem=parentItem)

    logging.info(f"Adding to cart with expansion. Product {product}")

    return call

def addToCart(cartId,parentProductName,productName,field='name',cartItems=None):

    restClient.logCall('getCartItems12')
    if cartItems == None:
        cartItems = getCartItems_api(cartId)

    parentItem = _getItemAtPath(cartItems,parentProductName,field)
    #parentItem = jsonUtil.getSibling(cartItems,'name',parentProductName,logValues=True)
    item = objectUtil.getSibling(parentItem,'name',productName)
   # if item == '':
   #     expanditems = 

    parentItem = copy.deepcopy(parentItem)
    parentItem['childProducts'] = ""

    execute = executeActions(item,parentItem=parentItem)
    return execute

def expand_action(cartItems,sibling,field='ProductCode'):    
    sibblingItem = objectUtil.getSiblingWhere(cartItems,field,sibling,'itemType','lineItem')
    if sibblingItem['object'] == None:
        sibblingItem = objectUtil.getSiblingWhere(cartItems,field,sibling,'itemType','productGroup')

    call = executeActions(sibblingItem['object']['actions'],'expanditems')
    logging.info(f"Expanding {sibling}")
    return call

def expand_api(cartId,itemId,productHierarchyPath):
    paramStr = stringify(locals(),exclude=[], initSeparator='?')

    action = f'/services/apexrest/vlocity_cmt/v2/cpq/carts/{cartId}/items/{itemId}/expand{paramStr}'

    call = restClient.callAPI( action, method="get")
    checkError()
    return call

def executeRestAction_api(rest):
    link = rest['link']
    params = rest['params']
    method = f"{rest['method']}".lower()

    call = restClient.callAPI(link, method=method, data=params)
    return call

def executeActions(actionsObject, actionName='addtocart', parentItem=None):
    if 'actions' in actionsObject:
        actions = actionsObject['actions']

    rest = actionsObject[actionName]['rest']
    if actionName == 'addtocart':
        rest['params']['items'][0]['parentRecord'] = {"records": [parentItem]}

    call = executeRestAction_api(rest)
    return call

def deleteCartItems_api(cartId, itemIds,parentRecord=None,hierarchy=None,lastRecordId=None,pagesize=None,includeAttachment=None,fields=None,query=None,filters=None,price=True,validate=True):
    paramStr = stringify(locals(),exclude=['cartid','itemIds'],initSeparator='&')
    paramStr = ''
    action =  f'/services/apexrest/vlocity_cmt/v2/cpq/carts/{cartId}/items?id={",".join(itemIds)}{paramStr}'
    if parentRecord == None:
        requestBody = {
            "items": [{
                "itemId": ",".join(itemIds)
            }]#,
        # "price": True,
        # "validate": True
        }
    if parentRecord != None:
        requestBody = {
            "methodName": "deleteCartsItems",
            "id": itemIds,
            "filters": filters,
            "fields": fields,
            "cartId": cartId,
            "price": price,
            "validate": validate,
            "includeAttachment": False,
            "pagesize": 10,
            "lastRecordId": None,
            "hierarchy": -1,
            "query": None,
            "items": [
                {
                    "parentRecord": {
                        "records": [
                            parentRecord
                        ]
                    }
                }
            ]
        }
   # parentRecord['noResponseNeeded'] = True  #This does not work

    call = restClient.callAPI(action,method="delete",data=requestBody)
    checkError()
    restClient.glog().info(f"Deleted cart Item {itemIds} for cart {cartId}")

    return call

def getCartPromotions(cartF,getPromotionsAppliedToCart=False,query=None,include=None,includePenalties=None,ruleType=None,ruleEvaluationInput=None,filters=None,category=None,fields=None,pagesize=None,commitmentDateFilter=None,appliedPromoStatusFilter=None,onlyOne=False):
    cartId = queryM.queryIdF('Order',cartF)
    promos = getCartPromotions_api(cartId=cartId,
                            getPromotionsAppliedToCart=getPromotionsAppliedToCart,
                            query=query,
                            include=include,
                            includePenalties=includePenalties,
                            ruleType=ruleType,
                            ruleEvaluationInput=ruleEvaluationInput,
                            filters=filters,
                            category=category,
                            fields=fields,
                            pagesize=pagesize,
                            commitmentDateFilter=commitmentDateFilter,
                            appliedPromoStatusFilter=appliedPromoStatusFilter
                            )
    if promos == None:
        return None
    if onlyOne:
        promo = [promo  for promo in promos if promo['Name'] == query]
        if len(promo) == 0:
            return None
        return promo[0]
    return promos

def getCartPromotions_api(cartId,getPromotionsAppliedToCart=False,query=None,include=None,includePenalties=None,ruleType=None,ruleEvaluationInput=None,filters=None,category=None,fields=None,pagesize=None,commitmentDateFilter=None,appliedPromoStatusFilter=None):
    subaction =  "getPromotionsAppliedToCart" if getPromotionsAppliedToCart == True else None

    paramStr = stringify(locals(),exclude=['cartId','getPromotionsAppliedToCart'])
    action = f'/services/apexrest/vlocity_cmt/v2/cpq/carts/{cartId}/promotions{paramStr}'

    call = restClient.callAPI(action)
    checkError()
    if call['totalSize'] == 0:
        return None
    return call['records']

def postCartsPromoItems_api(cartId, promotionId,price=True,validate=True):
    action = f'/services/apexrest/vlocity_cmt/v2/cpq/carts/{cartId}/promotions?cartId={cartId}&id={promotionId}'

    data = {
        "methodName": "postCartsPromoItems",
        "items": [{
            "itemId": promotionId
        }],
        "promotionId": promotionId,
        "cartId": cartId,
        "price":price,
        "validate":validate
    }
    call = restClient.callAPI(action, method='post', data=data)
    checkError()
    restClient.glog().info(f"Promo {promotionId} added to cart {cartId}")
    return call

def deleteCartPromotion(cartF, promotion,price=True,validate=True):
    cartId = queryM.queryIdF('Order',cartF)

    if 'Id' in promotion:
        promotion = promotion['Id']
    if 'value' in promotion:
        promotion = promotion['value']
    return deleteCartPromotion_api(cartId=cartId, promotionId=promotion,price=price,validate=validate)

def deleteCartPromotion_api(cartId, promotionId,price=True,validate=True):
    action = f'/services/apexrest/vlocity_cmt/v2/cpq/carts/{cartId}/promotions?cartId={cartId}&id={promotionId}'
    data = {
        "id": promotionId,
        "cartId": cartId,
        "methodName": "deleteAppliedPromoItems",
        "price": price,
        "validate": validate
    }
    call = restClient.callAPI(action, method='delete', data=data)
    checkError()
    return call

def getCartProducts(cartF, maxProdListHierarchy=None, query=None, filters=None,getCartProducts=None,lastRecordId=None,includeAttachment=None,offsetSize=None,fields=None,pagesize=None,attributes=None,includeAttributes=None,includeIneligible=True):
    cartId = query.queryIdF('Order',cartF)

def getCartProducts_api(cartId, maxProdListHierarchy=None, query=None, filters=None,getCartProducts=None,lastRecordId=None,includeAttachment=None,offsetSize=None,fields=None,pagesize=None,attributes=None,includeAttributes=None,includeIneligible=True):
    """
    - cartId:  Cart Id (Salesforce Id).Opportunity, Quote or Order Id. - Required. String
    - maxProdListHierarchy: hierarchy depth returned for list of products. Integer. 
    - lastRecordId: The last record ID from the previous search result, if available. string.
    - query: search string. 
    - includeAttachment: Whether product attachments are returned. boolean
    - filters: Filter field values. string
    - offsetSize: Offset from which to start reading products, for pagination. Integet.
    - fields: List of fields to return in the response, separated by commas. 
    - pagesize: Number of records to be returned. integer.
    - attributes: Attribute filters. 
    - includeAttributes: Specifies whether to return a list of attributes and their values for the product. boolean
    """
    paramStr = stringify(locals(),exclude=['cartId'])

    cartId = utils.Id(cartId)
    action = f'/services/apexrest/vlocity_cmt/v2/cpq/carts/{cartId}/products{paramStr}'

    call = restClient.callAPI(action)

    checkError()
   # if call['messages'][0]['message'] == 'No Results Found.':
   #     return None
    return call['records']

def checkOut_api(cartId,validateSubmittedXLI=True,assetizeFullBundlePerRoot=True,checkOrderStatus=True, provisioningStatus='Active',skipCheckoutValidation=True,waitActivation=False,asyncCall = False):
    cartId = utils.Id(cartId)

    data = {
        "methodName": "checkout",
        "cartId": cartId,
        "ContextId": cartId,
        #     "objectTypeToIdsToClone":"OrderItem:8021N000008NG5t_8021N000008NG5s",
        "validateSubmittedXLI": validateSubmittedXLI,
        "assetizeFullBundlePerRoot": assetizeFullBundlePerRoot,
        "checkOrderStatus": checkOrderStatus,
        "provisioningStatus": provisioningStatus,
        "skipCheckoutValidation": skipCheckoutValidation
    }

    action = f'/services/apexrest/vlocity_cmt/v2/cpq/carts/{cartId}/items/checkout'
    if asyncCall == True:
        action = f'/services/apexrest/AsyncCPQAppHandler/v1/checkout'
    call = restClient.callAPI(
        action,
        method="post",
        data=data)
    checkError()
    logging.info(f'Cart Check-out {cartId}')

    assert(False)
   # if waitActivation == True:
   #   orderUtils.waitforOrderActivation(cartId)
    
    return call

#{"methodName":"postCartsDiscounts","cartId":"8013O000003iwBIQAY","items":{"records":[{"Allocation":{"selectedDiscountType":{"value":"Order","name":"Order","$$hashKey":"object:3829"},"types":[{"value":"Account","name":"Account","$$hashKey":"object:3828"},{"value":"Order","name":"Order","$$hashKey":"object:3829"},{"value":"Contract","name":"Contract","$$hashKey":"object:3830"}],"label":"Discount Type"},"Category":{"actions":{"rest":{"params":{},"method":null,"link":null},"remote":{"params":{"methodName":"getCatalogHierarchy","ContextId":"8013O000003iwBIQAY","includeAllCatalogs":true,"cartId":"8013O000003iwBIQAY"}},"client":{"records":[],"params":{"methodName":"getCatalogHierarchy","ContextId":"8013O000003iwBIQAY","includeAllCatalogs":true,"cartId":"8013O000003iwBIQAY"}}},"categories":[],"value":"","label":"Category"},"Product":{"actions":{"rest":{"params":{},"method":null,"link":null},"remote":{"params":{"methodName":"getCartsProducts","cartId":"8013O000003iwBIQAY"}},"client":{"records":[],"params":{"methodName":"getCartsProducts","cartId":"8013O000003iwBIQAY"}}},"products":[],"value":"","label":"Product"},"Discount":{"discounts":[{"types":[{"method":"Percent","detailType":"Discount","value":"%"},{"method":"Absolute","detailType":"Discount","value":"€"}],"value":-1,"actions":{"rest":{"params":{},"method":null,"link":null},"remote":{"params":{"methodName":"getListsOfValues","listkeys":"TimePlans,TimePolicies"}},"client":{"records":[],"params":{}}},"selectedAdjustmentMethod":{"value":"%"},"chargeType":"Recurring","selectedTimePlan":null,"selectedTimePolicy":null,"label":"","$$hashKey":"object:3813"},{"types":[{"method":"Percent","detailType":"Discount","value":"%"},{"method":"Absolute","detailType":"Discount","value":"€"}],"value":-1,"selectedAdjustmentMethod":{"value":"%"},"chargeType":"One-time","label":"","$$hashKey":"object:3812"}],"label":"Discount","value":""},"Duration unit":{"label":"Duration unit","value":"Month"},"All items in cart":{"label":"All items in cart","value":true},"End date":{"label":"End Date","value":null},"Active period":{"label":"Active Period","value":null},"Discount Name":{"label":"Discount Name","value":"zzz1"},"timePlanList":[{"valuekey":"a5v3O0000002TGZQA2","label":"84de63e5-4eb4-51de-74c4-90b43d2e3bd6"},{"valuekey":"a5v3O0000002TGjQAM","label":"TP-3M"},{"valuekey":"a5v3O0000002TGoQAM","label":"TP1"},{"valuekey":"a5v3O0000002NlmQAE","label":"TP-24M"},{"valuekey":"a5v3O0000002NlnQAE","label":"TP-12M"},{"valuekey":"a5v3O0000002NloQAE","label":"TP-1M"},{"valuekey":"a5v3O0000002NlpQAE","label":"TP-6M"}],"Description":{"value":"this is the description"}}]}}
def postCartDiscount():
    print()



import os
import errno
from . import utils

def abspath(filepath):
    return os.path.abspath(filepath)

def addExtension(filename,extension):
    if '.' not in extension:
        extension = '.' + extension
    if not filename.lower().endswith(extension):
        filename = filename +extension
    return filename

def _createDirectories(filename):
    dirname =os.path.dirname(filename)
    if dirname == '':
        return True #local file
    if not os.path.exists(os.path.dirname(filename)):
        try:
            os.makedirs(os.path.dirname(filename))
            return True
        except OSError as exc: # Guard against race condition
            if exc.errno != errno.EEXIST:
                utils.raiseException('OSError',f"Error creating directory {dirname}  {exc.errno}")
    return False

#def getFilePath(folder,filename):

def writeFileinFolder(folder,filename,data):
    return write(f"{folder}/{filename}",'.txt',data)

def delete(filename):
    os.remove(filename)


def write(filepath,data,mode="w"):
    _createDirectories(filepath)
    f = open(filepath,mode)
    f.write(data)
    f.close()

    return os.path.abspath(filepath)

def exists(filepath):
    return os.path.exists(filepath)

def read(filepath):
    f = open(filepath,"r")
    data = f.read()
    f.close()
    return data

def openFile(filepath,mode="w"):
    f = open(filepath,mode)
    return f
def write_line(f,line):
    f.write(line+ "\n")
    return f
def closeFile(f):
    f.close()

def emptyDirectory(directory):
    for f in os.listdir(directory):
        if os.path.isfile(os.path.join(directory, f)):
            os.remove(os.path.join(directory, f))


def writeDictionaty(filepath,extension,dictionary,firstLine=False,mode='a'):
  #  filepath = _checkExtension(filepath,extension)

    if os.path.isfile(filepath) == False:  #If theere is no file, always create header
        firstLine = True
    else:
        firstLine = False

    if firstLine == True:
        joined = ",".join(dictionary)+ "\n"
        #print(joined)
        write(filepath,extension,joined,mode=mode)

    values = list(dictionary.values())
    str_values = map(str,values)

    joined = ",".join(str_values) + "\n"
    write(filepath,extension,joined,'a')

